import os

PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(PROJECT_DIR)
SECRET_KEY = os.getenv("HUB20_SECRET_KEY")

DEBUG = "HUB20_DEBUG" in os.environ


ALLOWED_HOSTS = os.getenv("HUB20_ALLOWED_HOSTS", ["*"])

# Application definition

DJANGO_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.humanize",
    "django.contrib.admin",
    "django.contrib.postgres",
]

THIRD_PARTY_APPS = [
    "allauth",
    "allauth.account",
    "allauth.socialaccount",
    "channels",
    "corsheaders",
    "dj_rest_auth",
    "dj_rest_auth.registration",
    "django_celery_beat",
    "django_filters",
    "djmoney",
    "drf_link_header_pagination",
    "drf_yasg",
    "rest_framework",
    "rest_framework.authtoken",
    "taggit",
]

PROJECT_APPS = [
    "hub20.apps.blockchain",
    "hub20.apps.wallet",
    "hub20.apps.ethereum_money",
    "hub20.apps.raiden",
    "hub20.apps.core",
]

INSTALLED_APPS = DJANGO_APPS + THIRD_PARTY_APPS + PROJECT_APPS

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

APPEND_SLASHES = False
ROOT_URLCONF = os.getenv("HUB20_URLCONF_MODULE", "hub20.api.urls")
ASGI_APPLICATION = "hub20.api.asgi.application"
SERVE_OPENAPI_URLS = "HUB20_SERVE_OPENAPI_URLS" in os.environ or DEBUG

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(PROJECT_DIR, "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]


# Database
# https://docs.djangoproject.com/en/1.9/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",  # PostgreSQL is required
        "HOST": os.getenv("HUB20_DATABASE_HOST"),
        "PORT": os.getenv("HUB20_DATABASE_PORT", 5432),
        "NAME": os.getenv("HUB20_DATABASE_NAME"),
        "USER": os.getenv("HUB20_DATABASE_USER"),
        "PASSWORD": os.getenv("HUB20_DATABASE_PASSWORD"),
    }
}

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"


# Both Channel Layer and the cache system will require a running redis
# server, so the only thing you can change here is the URL (host/port/db)


# Channel Configurations
CHANNEL_LAYER_BACKEND = "channels_redis.core.RedisChannelLayer"
CHANNEL_LAYER_URL = os.getenv("HUB20_CHANNEL_LAYER_URL", "redis://redis:6379/2")
CHANNEL_LAYERS = {
    "default": {
        "BACKEND": CHANNEL_LAYER_BACKEND,
        "CONFIG": {"hosts": (CHANNEL_LAYER_URL,)},
    }
}

# Cache configuration
CACHE_BACKEND = "django_redis.cache.RedisCache"
CACHE_LOCATION = os.getenv("HUB20_CACHE_LOCATION", "redis://redis:6379/1")
CACHE_OPTIONS = {"CLIENT_CLASS": "django_redis.client.DefaultClient"}
CACHES = {
    "default": {"BACKEND": CACHE_BACKEND, "LOCATION": CACHE_LOCATION, "OPTIONS": CACHE_OPTIONS}
}


# Celery can use redis and many other systems as its task broker (e.g,
# RabbitMQ). Given that we already have a running redis for cache and
# channel layers, we use the same server in the default configuration

# Celery configuration
CELERY_BROKER_URL = os.getenv("HUB20_BROKER_URL", "redis://redis:6379/0")


# This needs to be set up if you don't have a front-end proxy server
CORS_ORIGIN_ALLOW_ALL = "HUB20_CORS_HEADERS_ENABLED" in os.environ

# This ensures that requests are seen as secure when the proxy sets the header
SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")

# Password validation
# https://docs.djangoproject.com/en/1.9/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {"NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"},
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


# Email
DEFAULT_FROM_EMAIL = os.getenv("HUB20_EMAIL_MAILER_ADDRESS")
EMAIL_BACKEND = os.getenv("HUB20_EMAIL_BACKEND")
EMAIL_HOST = os.getenv("HUB20_EMAIL_HOST")
EMAIL_PORT = os.getenv("HUB20_EMAIL_PORT")
EMAIL_HOST_USER = os.getenv("HUB20_EMAIL_SMTP_USERNAME")
EMAIL_HOST_PASSWORD = os.getenv("HUB20_EMAIL_SMTP_PASSWORD")
EMAIL_TIMEOUT = os.getenv("HUB20_EMAIL_TIMEOUT", 5)


# Internationalization
# https://docs.djangoproject.com/en/1.9/topics/i18n/

LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = True
USE_TZ = True


SITE_ID = 1
HUB20_SITE_DOMAIN = os.getenv("HUB20_SITE_DOMAIN")


# Rest Framework settings
REST_RENDERERS = ("BrowsableAPIRenderer", "JSONRenderer") if DEBUG else ("JSONRenderer",)
REST_FRAMEWORK = {
    "DEFAULT_FILTER_BACKENDS": ("django_filters.rest_framework.DjangoFilterBackend",),
    "DEFAULT_RENDERER_CLASSES": [
        f"rest_framework.renderers.{renderer}" for renderer in REST_RENDERERS
    ],
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework.authentication.TokenAuthentication",
        "rest_framework.authentication.SessionAuthentication",
    ],
    "DEFAULT_PAGINATION_CLASS": "drf_link_header_pagination.LinkHeaderPagination",
    "PAGE_SIZE": 250,
}

# REST Auth
REST_AUTH_SERIALIZERS = {
    "USER_DETAILS_SERIALIZER": "hub20.api.serializers.UserProfileSerializer",
}


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.9/howto/static-files/

STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
)

STATIC_URL = "/static/"
STATIC_ROOT = os.getenv("HUB20_STATIC_ROOT", os.path.abspath(os.path.join(BASE_DIR, "static")))

ADMIN_USERNAME = os.getenv("HUB20_ADMIN_USERNAME", "admin")


# Configuration of authentication/signup/registration via django-allauth/invitations
LOGIN_URL = "login"
LOGIN_REDIRECT_URL = "rest_user_details"
ACCOUNT_LOGIN_ON_PASSWORD_RESET = True
ACCOUNT_EMAIL_REQUIRED = False
ACCOUNT_EMAIL_VERIFICATION = "none"
ACCOUNT_UNIQUE_EMAIL = True
ACCOUNT_USERNAME_BLACKLIST = [
    "about",
    "access",
    "account",
    "accounts",
    "add",
    "address",
    "adm",
    "admin",
    "administration",
    "adult",
    "advertising",
    "affiliate",
    "affiliates",
    "ajax",
    "analytics",
    "android",
    "anon",
    "anonymous",
    "api",
    "app",
    "apps",
    "archive",
    "atom",
    "auth",
    "authentication",
    "avatar",
    "backup",
    "banner",
    "banners",
    "bin",
    "billing",
    "blog",
    "blogs",
    "board",
    "bot",
    "bots",
    "business",
    "chat",
    "cache",
    "cadastro",
    "calendar",
    "campaign",
    "careers",
    "cgi",
    "client",
    "cliente",
    "code",
    "comercial",
    "compare",
    "config",
    "connect",
    "contact",
    "contest",
    "create",
    "code",
    "compras",
    "css",
    "dashboard",
    "data",
    "db",
    "design",
    "delete",
    "demo",
    "design",
    "designer",
    "dev",
    "devel",
    "dir",
    "directory",
    "doc",
    "docs",
    "domain",
    "download",
    "downloads",
    "edit",
    "editor",
    "email",
    "ecommerce",
    "forum",
    "forums",
    "faq",
    "favorite",
    "feed",
    "feedback",
    "flog",
    "follow",
    "file",
    "files",
    "free",
    "ftp",
    "gadget",
    "gadgets",
    "games",
    "guest",
    "group",
    "groups",
    "help",
    "home",
    "homepage",
    "host",
    "hosting",
    "hostname",
    "html",
    "http",
    "httpd",
    "https",
    "hpg",
    "info",
    "information",
    "image",
    "img",
    "images",
    "imap",
    "index",
    "invite",
    "intranet",
    "indice",
    "ipad",
    "iphone",
    "irc",
    "java",
    "javascript",
    "job",
    "jobs",
    "js",
    "knowledgebase",
    "log",
    "login",
    "logs",
    "logout",
    "list",
    "lists",
    "mail",
    "mail1",
    "mail2",
    "mail3",
    "mail4",
    "mail5",
    "mailer",
    "mailing",
    "mx",
    "manager",
    "marketing",
    "master",
    "me",
    "media",
    "message",
    "microblog",
    "microblogs",
    "mine",
    "mp3",
    "msg",
    "msn",
    "mysql",
    "messenger",
    "mob",
    "mobile",
    "movie",
    "movies",
    "music",
    "musicas",
    "my",
    "name",
    "named",
    "net",
    "network",
    "new",
    "news",
    "newsletter",
    "nick",
    "nickname",
    "notes",
    "noticias",
    "ns",
    "ns1",
    "ns2",
    "ns3",
    "ns4",
    "old",
    "online",
    "operator",
    "order",
    "orders",
    "page",
    "pager",
    "pages",
    "panel",
    "password",
    "perl",
    "pic",
    "pics",
    "photo",
    "photos",
    "photoalbum",
    "php",
    "plugin",
    "plugins",
    "pop",
    "pop3",
    "post",
    "postmaster",
    "postfix",
    "posts",
    "profile",
    "project",
    "projects",
    "promo",
    "pub",
    "public",
    "python",
    "random",
    "register",
    "registration",
    "root",
    "ruby",
    "rss",
    "sale",
    "sales",
    "sample",
    "samples",
    "script",
    "scripts",
    "secure",
    "send",
    "service",
    "shop",
    "sql",
    "signup",
    "signin",
    "search",
    "security",
    "settings",
    "setting",
    "setup",
    "site",
    "sites",
    "sitemap",
    "smtp",
    "soporte",
    "ssh",
    "stage",
    "staging",
    "start",
    "subscribe",
    "subdomain",
    "suporte",
    "support",
    "stat",
    "static",
    "stats",
    "status",
    "store",
    "stores",
    "system",
    "tablet",
    "tablets",
    "tech",
    "telnet",
    "test",
    "test1",
    "test2",
    "test3",
    "teste",
    "tests",
    "theme",
    "themes",
    "tmp",
    "todo",
    "task",
    "tasks",
    "tools",
    "tv",
    "talk",
    "update",
    "upload",
    "url",
    "user",
    "username",
    "usuario",
    "usage",
    "vendas",
    "video",
    "videos",
    "visitor",
    "win",
    "ww",
    "www",
    "www1",
    "www2",
    "www3",
    "www4",
    "www5",
    "www6",
    "www7",
    "wwww",
    "wws",
    "wwws",
    "web",
    "webmail",
    "website",
    "websites",
    "webmaster",
    "workshop",
    "xxx",
    "xpg",
    "you",
    "yourname",
    "yourusername",
    "yoursite",
    "yourdomain",
]

# Wallet model configuration
WALLET_MODEL = os.getenv("HUB20_WALLET_MODEL", "wallet.KeystoreAccount")

ETHEREUM_HD_WALLET_MNEMONIC = os.getenv("HUB20_ETHEREUM_HD_WALLET_MNEMONIC")
ETHEREUM_HD_WALLET_ROOT_KEY = os.getenv("HUB20_ETHEREUM_HD_WALLET_ROOT_KEY")


# Logging Configuration
LOG_FILE = os.getenv("HUB20_SITE_LOG_FILE")
LOG_LEVEL = os.getenv("HUB20_LOG_LEVEL", "DEBUG" if DEBUG else "INFO")


LOGGING_HANDLERS = {
    "null": {"level": "DEBUG", "class": "logging.NullHandler"},
    "console": {"level": LOG_LEVEL, "class": "logging.StreamHandler", "formatter": "verbose"},
}

if LOG_FILE:
    LOGGING_HANDLERS["file"] = {
        "level": LOG_LEVEL,
        "class": os.getenv("HUB20_LOG_FILE_HANDLER", "logging.handlers.RotatingFileHandler"),
        "formatter": "verbose",
        "filename": LOG_FILE,
        "maxBytes": int(os.getenv("HUB20_LOG_FILE_MAX_FILE_SIZE", 16 * 1024 * 1024)),
        "backupCount": int(os.getenv("HUB20_LOG_FILE_BACKUP", 3)),
    }

LOGGING_HANDLER_METHODS = ["console", "file"] if "file" in LOGGING_HANDLERS else ["console"]

LOGGING = {
    "version": 1,
    "disable_existing_loggers": "HUB20_LOG_DISABLE_EXISTING_LOGGERS" in os.environ,
    "formatters": {
        "verbose": {
            "format": "%(asctime)s %(levelname)s:%(pathname)s %(process)d %(lineno)d %(message)s"
        },
        "simple": {"format": "%(levelname)s:%(module)s %(lineno)d %(message)s"},
    },
    "handlers": LOGGING_HANDLERS,
    "loggers": {
        "django": {"handlers": ["null"], "propagate": True, "level": "INFO"},
        "django.db.backends:": {
            "handlers": LOGGING_HANDLER_METHODS,
            "level": "ERROR",
            "propagate": False,
        },
        "django.request": {
            "handlers": LOGGING_HANDLER_METHODS,
            "level": "ERROR",
            "propagate": False,
        },
    },
}


for app in PROJECT_APPS:
    LOGGING["loggers"][app] = {
        "handlers": LOGGING_HANDLER_METHODS,
        "level": LOG_LEVEL,
        "propagate": False,
    }



from typing import Optional

from django import forms
from django.contrib import admin
from django.http import HttpRequest

from . import models
from .typing import EthereumAccount_T
from .validators import web3_url_validator


class Web3URLField(forms.URLField):
    default_validators = [web3_url_validator]


Web3ProviderForm = forms.modelform_factory(
    model=models.Web3Provider,
    fields=["chain", "url", "is_active", "connected", "synced"],
    field_classes={"url": Web3URLField},
)


@admin.register(models.Chain)
class ChainAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "is_mainnet", "provider")
    list_filter = ("is_mainnet",)
    readonly_fields = ("highest_block",)
    search_fields = ("name", "id")


@admin.register(models.Web3Provider)
class Web3ProviderAdmin(admin.ModelAdmin):
    form = Web3ProviderForm

    list_display = ("hostname", "chain", "is_active", "connected", "synced")
    list_filter = ("is_active", "connected", "synced")
    readonly_fields = ("connected", "synced")
    search_fields = ("url", "chain__name")


@admin.register(models.Explorer)
class BlockchainExplorerAdmin(admin.ModelAdmin):
    list_display = ("name", "url", "standard")
    list_filter = ("standard",)


@admin.register(models.BaseEthereumAccount)
class EthereumAccountAdmin(admin.ModelAdmin):
    list_display = ("address",)

    def has_add_permission(
        self, request: HttpRequest, obj: Optional[EthereumAccount_T] = None
    ) -> bool:
        return False

    def has_change_permission(
        self, request: HttpRequest, obj: Optional[EthereumAccount_T] = None
    ) -> bool:
        return False

    def has_delete_permission(
        self, request: HttpRequest, obj: Optional[EthereumAccount_T] = None
    ) -> bool:
        return False



import factory

from hub20.apps.blockchain.factories.providers import EthereumProvider
from hub20.apps.core import models
from hub20.apps.ethereum_money.factories import EthereumTokenValueModelFactory

from .base import UserFactory

factory.Faker.add_provider(EthereumProvider)


class TransferFactory(EthereumTokenValueModelFactory):
    sender = factory.SubFactory(UserFactory)

    class Meta:
        model = models.Transfer


class InternalTransferFactory(TransferFactory):
    receiver = factory.SubFactory(UserFactory)


class ExternalTransferFactory(TransferFactory):
    address = factory.Faker("ethereum_address")


__all__ = ["TransferFactory", "InternalTransferFactory", "ExternalTransferFactory"]



import factory
from factory import fuzzy

from hub20.apps.blockchain.factories import (
    EthereumProvider,
    SyncedChainFactory,
    TransactionFactory,
)
from hub20.apps.core import models
from hub20.apps.ethereum_money.factories import Erc20TokenFactory, ETHFactory
from hub20.apps.wallet.factories import EthereumAccountFactory

from .base import UserFactory

factory.Faker.add_provider(EthereumProvider)


class PaymentOrderFactory(factory.django.DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    amount = fuzzy.FuzzyDecimal(0, 10, precision=6)

    class Meta:
        abstract = False
        model = models.PaymentOrder


class EtherPaymentOrderFactory(PaymentOrderFactory):
    currency = factory.SubFactory(ETHFactory)

    class Meta:
        model = models.PaymentOrder


class Erc20TokenPaymentOrderFactory(PaymentOrderFactory):
    currency = factory.SubFactory(Erc20TokenFactory)

    class Meta:
        model = models.PaymentOrder


class EtherBlockchainPaymentRouteFactory(factory.django.DjangoModelFactory):
    deposit = factory.SubFactory(EtherPaymentOrderFactory)
    account = factory.SubFactory(EthereumAccountFactory)
    chain = factory.SubFactory(SyncedChainFactory)
    payment_window = factory.LazyAttribute(
        lambda obj: (obj.chain.highest_block, obj.chain.highest_block + 100)
    )

    class Meta:
        model = models.BlockchainPaymentRoute
        django_get_or_create = ("account",)


class Erc20TokenBlockchainPaymentRouteFactory(EtherBlockchainPaymentRouteFactory):
    deposit = factory.SubFactory(Erc20TokenPaymentOrderFactory)


class EtherBlockchainPaymentFactory(factory.django.DjangoModelFactory):
    route = factory.SubFactory(EtherBlockchainPaymentRouteFactory)
    transaction = factory.SubFactory(TransactionFactory)
    currency = factory.SelfAttribute("route.deposit.currency")
    amount = fuzzy.FuzzyDecimal(0, 10, precision=6)

    class Meta:
        model = models.BlockchainPayment


class Erc20TokenBlockchainPaymentFactory(EtherBlockchainPaymentFactory):
    route = factory.SubFactory(Erc20TokenBlockchainPaymentRouteFactory)


class EtherPaymentConfirmationFactory(factory.django.DjangoModelFactory):
    payment = factory.SubFactory(EtherBlockchainPaymentFactory)

    class Meta:
        model = models.PaymentConfirmation


class Erc20TokenPaymentConfirmationFactory(EtherPaymentConfirmationFactory):
    payment = factory.SubFactory(Erc20TokenBlockchainPaymentFactory)


__all__ = [
    "Erc20TokenBlockchainPaymentRouteFactory",
    "Erc20TokenBlockchainPaymentFactory",
    "Erc20TokenPaymentConfirmationFactory",
    "Erc20TokenPaymentOrderFactory",
    "EtherBlockchainPaymentRouteFactory",
    "EtherBlockchainPaymentFactory",
    "EtherPaymentConfirmationFactory",
    "EtherPaymentOrderFactory",
]



from rest_framework import serializers
from rest_framework.reverse import reverse
from taggit.serializers import TaggitSerializer, TagListSerializerField

from hub20.apps.blockchain.client import make_web3
from hub20.apps.blockchain.models import Chain, Web3Provider
from hub20.apps.blockchain.serializers import EthereumAddressField

from . import models
from .client import get_token_information, get_transfer_gas_estimate


class TokenValueField(serializers.DecimalField):
    def __init__(self, *args, **kw):
        kw.setdefault("max_digits", 32)
        kw.setdefault("decimal_places", 18)
        super().__init__(*args, **kw)


class HyperlinkedTokenMixin:
    def get_queryset(self):
        return models.EthereumToken.tradeable.all()

    def get_url(self, obj, view_name, request, format):
        url_kwargs = {"chain_id": obj.chain_id, "address": obj.address}
        return reverse(view_name, kwargs=url_kwargs, request=request, format=format)

    def get_object(self, view_name, view_args, view_kwargs):
        lookup_kwargs = {"chain_id": view_kwargs["chain_id"], "address": view_kwargs["address"]}
        queryset = self.get_queryset()
        return queryset.get(**lookup_kwargs)


class HyperlinkedRelatedTokenField(HyperlinkedTokenMixin, serializers.HyperlinkedRelatedField):
    queryset = models.EthereumToken.tradeable.all()
    view_name = "token-detail"

    def get_attribute(self, instance):
        return getattr(instance, self.source)


class HyperlinkedTokenIdentityField(serializers.HyperlinkedIdentityField):
    def get_url(self, obj, view_name, request, format):
        url_kwargs = {"chain_id": obj.chain_id, "address": obj.address}
        return reverse(view_name, kwargs=url_kwargs, request=request, format=format)


class EthereumTokenSerializer(serializers.ModelSerializer):
    address = EthereumAddressField()
    chain_id = serializers.PrimaryKeyRelatedField(queryset=Chain.active.all())

    def validate(self, data):
        chain = data.pop("chain_id")
        address = data.pop("address")

        if chain.tokens.filter(address=address).exists():
            raise serializers.ValidationError(f"Token {address} already registered")

        provider = Web3Provider.available.filter(chain_id=chain.id).first()
        if not provider:
            raise serializers.ValidationError(f"Could not query {chain.id} for token information")

        w3 = make_web3(provider=provider)
        try:
            token_data = get_token_information(w3, address=address)
        except Exception:
            raise serializers.ValidationError(
                f"Could not get token information for address {address}"
            )

        data.update(token_data)
        data.update({"chain": chain, "address": address})

        try:
            token = models.EthereumToken(**data)
            get_transfer_gas_estimate(w3=w3, token=token)
        except Exception:
            raise serializers.ValidationError(
                f"Could not verify {address} as a tradeable ERC20 token"
            )

        return data

    class Meta:
        model = models.EthereumToken
        fields = (
            "symbol",
            "name",
            "address",
            "chain_id",
            "decimals",
            "logoURI",
        )
        read_only_fields = (
            "symbol",
            "name",
            "decimals",
            "logoURI",
        )


class EthereumTokenAmountSerializer(serializers.ModelSerializer):
    token = HyperlinkedRelatedTokenField(source="currency")
    amount = TokenValueField()


class HyperlinkedEthereumTokenSerializer(EthereumTokenSerializer):
    url = HyperlinkedTokenIdentityField(view_name="token-detail")

    class Meta:
        model = models.EthereumToken
        fields = ("url",) + EthereumTokenSerializer.Meta.fields
        read_only_fields = ("url",) + EthereumTokenSerializer.Meta.read_only_fields


class TokenInfoSerializer(serializers.ModelSerializer):
    wrapped_by = HyperlinkedTokenIdentityField(view_name="token-detail", read_only=True, many=True)
    wraps = HyperlinkedRelatedTokenField()

    class Meta:
        model = models.EthereumToken
        fields = read_only_fields = (
            "name",
            "symbol",
            "wrapped_by",
            "wraps",
            "is_stable",
            "tracks_currency",
        )


class BaseTokenListSerializer(serializers.ModelSerializer):
    tokens = HyperlinkedRelatedTokenField(many=True)


class TokenListSerializer(TaggitSerializer, BaseTokenListSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="tokenlist-detail")
    source = serializers.URLField(source="url")
    keywords = TagListSerializerField(read_only=True)

    class Meta:
        model = models.TokenList
        fields = read_only_fields = (
            "url",
            "id",
            "version",
            "source",
            "name",
            "description",
            "tokens",
            "keywords",
        )


class UserTokenListSerializer(BaseTokenListSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="user-tokenlist-detail")
    keywords = TagListSerializerField()

    def create(self, validated_data):
        request = self.context.get("request")
        tokens = validated_data.pop("tokens")
        user_token_list = request.user.token_lists.create(**validated_data)
        user_token_list.tokens.set(tokens)
        return user_token_list

    class Meta:
        model = models.UserTokenList
        fields = ("url", "id", "name", "description", "tokens", "keywords")
        read_only_fields = ("url", "id")


class TokenListImportSerializer(serializers.ModelSerializer):
    url = serializers.URLField()

    def validate(self, data):
        url = data["url"]
        try:
            token_list_data = models.TokenList.fetch(url)
        except ValueError as exc:
            raise serializers.ValidationError(exc)

        data.update({"token_list_data": token_list_data})
        return data

    class Meta:
        model = models.TokenList
        fields = ("url", "description")


class UserTokenListCloneSerializer(UserTokenListSerializer):
    token_list = serializers.HyperlinkedRelatedField(
        view_name="tokenlist-detail", queryset=models.TokenList.objects.all(), write_only=True
    )

    def create(self, validated_data):
        request = self.context["request"]
        return models.UserTokenList.clone(
            user=request.user, token_list=validated_data["token_list"]
        )

    class Meta:
        model = models.UserTokenList
        fields = UserTokenListSerializer.Meta.fields + ("token_list",)
        read_only_fields = ("url", "name", "description", "tokens", "keywords")



class RaidenOperationError(Exception):
    pass


class RaidenMissingPrecondition(Exception):
    pass


class RaidenConnectionError(Exception):
    pass


class RaidenPaymentError(Exception):
    def __init__(self, error_code, message):
        self.error_code = error_code
        self.message = message

    def __str__(self):
        return f"Payment Error: {self.message} ({self.error_code})"


class InsufficientBalanceError(Exception):
    pass



from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.reverse import reverse_lazy
from rest_framework.views import APIView

from . import VERSION


class IndexView(APIView):
    """
    Hub20 Root Endpoint. Provides links to other endpoints and informs version for clients.
    """

    permission_classes = (AllowAny,)

    def get(self, request, **kw):
        return Response(
            {
                "current_user_url": reverse_lazy("rest_user_details", request=request),
                "blockchains_url": reverse_lazy("blockchain:chain-list", request=request),
                "accounting_report_url": reverse_lazy("accounting-report", request=request),
                "tokens_url": reverse_lazy("token-list", request=request),
                "users_url": reverse_lazy("users-list", request=request),
                "version": VERSION,
            }
        )



import factory
from factory import fuzzy

from hub20.apps.blockchain.factories import (
    EthereumProvider,
    SyncedChainFactory,
    TransactionDataFactory,
    TransactionFactory,
)
from hub20.apps.ethereum_money.client import encode_transfer_data
from hub20.apps.ethereum_money.models import (
    EthereumToken,
    EthereumTokenAmount,
    StableTokenPair,
    TokenList,
    UserTokenList,
    WrappedToken,
)

factory.Faker.add_provider(EthereumProvider)


class EthereumCurrencyFactory(factory.django.DjangoModelFactory):
    chain = factory.SubFactory(SyncedChainFactory)
    is_listed = True


class ETHFactory(EthereumCurrencyFactory):
    name = "Ether"
    symbol = "ETH"
    address = EthereumToken.NULL_ADDRESS

    class Meta:
        model = EthereumToken
        django_get_or_create = ("chain", "name", "address")


class Erc20TokenFactory(EthereumCurrencyFactory):
    name = factory.Sequence(lambda n: f"ERC20 Token #{n:03}")
    symbol = factory.Sequence(lambda n: f"TOK#{n:03}")
    address = factory.Faker("ethereum_address")

    class Meta:
        model = EthereumToken


class WrappedEtherFactory(factory.django.DjangoModelFactory):
    wrapper = factory.SubFactory(Erc20TokenFactory)
    wrapped = factory.SubFactory(ETHFactory)

    class Meta:
        model = WrappedToken


class WrappedTokenFactory(factory.django.DjangoModelFactory):
    wrapper = factory.SubFactory(Erc20TokenFactory)
    wrapped = factory.SubFactory(Erc20TokenFactory)

    class Meta:
        model = WrappedToken


class StableTokenFactory(factory.django.DjangoModelFactory):
    token = factory.SubFactory(Erc20TokenFactory)
    currency = factory.Iterator(["USD", "EUR", "GBP"])

    class Meta:
        model = StableTokenPair


class EthereumTokenValueModelFactory(factory.django.DjangoModelFactory):
    amount = fuzzy.FuzzyDecimal(0, 10, precision=6)
    currency = factory.SubFactory(ETHFactory)


class Erc20TokenValueModelFactory(factory.django.DjangoModelFactory):
    amount = fuzzy.FuzzyDecimal(0, 1000, precision=8)
    currency = factory.SubFactory(Erc20TokenFactory)


class Erc20TokenAmountFactory(factory.Factory):
    amount = fuzzy.FuzzyDecimal(0, 10, precision=6)
    currency = factory.SubFactory(Erc20TokenFactory)

    class Meta:
        model = EthereumTokenAmount


class EtherAmountFactory(factory.Factory):
    amount = fuzzy.FuzzyDecimal(0, 10, precision=6)
    currency = factory.SubFactory(ETHFactory)

    class Meta:
        model = EthereumTokenAmount


class Erc20TransactionDataFactory(TransactionDataFactory):
    data = factory.LazyAttribute(
        lambda obj: {
            "from": obj.from_address,
            "to": obj.to_address,
            "status": obj.status,
            "blockNumber": obj.block_number,
            "blockHash": obj.block_hash,
            "gasUsed": obj.gas_used,
            "logs": [encode_transfer_data(obj.to_address, obj.amount)],
        }
    )

    class Params:
        gas_used = factory.fuzzy.FuzzyInteger(50000, 200000)
        amount = factory.SubFactory(Erc20TokenAmountFactory)


class Erc20TransferFactory(TransactionFactory):
    class Params:
        recipient = factory.Faker("ethereum_address")
        amount = factory.SubFactory(Erc20TokenAmountFactory)
        to_address = factory.LazyAttribute(lambda obj: obj.amount.currency.address)


class BaseTokenListFactory(factory.django.DjangoModelFactory):
    name = factory.Sequence(lambda n: f"Token List #{n:02}")
    description = factory.Sequence(lambda n: f"Description for token list #{n:02}")

    @factory.post_generation
    def tokens(self, create, tokens, **kw):
        if not create:
            return

        if tokens:
            for token in tokens:
                self.tokens.add(token)


class TokenListFactory(BaseTokenListFactory):
    url = factory.Sequence(lambda n: f"http://tokenlist{n:02}.example.com")

    class Meta:
        model = TokenList


class UserTokenListFactory(BaseTokenListFactory):
    class Meta:
        model = UserTokenList


__all__ = [
    "ETHFactory",
    "Erc20TokenFactory",
    "EthereumTokenValueModelFactory",
    "Erc20TokenValueModelFactory",
    "Erc20TokenAmountFactory",
    "EtherAmountFactory",
    "Erc20TransferFactory",
    "TokenListFactory",
    "WrappedEtherFactory",
    "WrappedTokenFactory",
    "StableTokenFactory",
    "UserTokenListFactory",
]



import logging

from django.core.management.base import BaseCommand

from hub20.apps.blockchain.client import get_web3, inspect_web3
from hub20.apps.blockchain.models import Chain, Web3Provider

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Sets the web3 rpc node used to interact with the blockchain"

    def add_arguments(self, parser):
        parser.add_argument("urls", metavar="N", nargs="+", type=str)

    def handle(self, *args, **options):

        for url in options["urls"]:
            w3 = get_web3(provider_url=url)
            try:
                chain_id = w3.eth.chain_id
                chain = Chain.objects.filter(id=chain_id).first()
                if not chain:
                    logger.warn(
                        f"Node at {url} is connected to network {chain_id}, which we do not know"
                    )
                    continue
                configuration = inspect_web3(w3)
                provider, _ = Web3Provider.objects.update_or_create(
                    chain=chain, url=url, defaults=configuration.dict()
                )
                provider.activate()
            except Exception as exc:
                logger.info(f"Error when connecting to {url}: {exc}")



import logging

from django.db.models.signals import post_save
from django.dispatch import receiver

from hub20.apps.core.models import Store, StoreRSAKeyPair

logger = logging.getLogger(__name__)


@receiver(post_save, sender=Store)
def on_store_created_generate_key_pair(sender, **kw):
    store = kw["instance"]
    if kw["created"]:
        StoreRSAKeyPair.generate(store)


__all__ = [
    "on_store_created_generate_key_pair",
]



from typing import List

import factory
import factory.fuzzy
from django.utils import timezone
from faker import Faker

from ..app_settings import START_BLOCK_NUMBER
from ..models import (
    BaseEthereumAccount,
    Block,
    Chain,
    NativeToken,
    Transaction,
    TransactionDataRecord,
    Web3Provider,
)
from .providers import EthereumProvider

factory.Faker.add_provider(EthereumProvider)


TEST_CHAIN_ID = 2
FAKER = Faker()
FAKER.add_provider(EthereumProvider)


def find_parent_by_block_number(block):
    return Block.objects.filter(chain=block.chain, number=block.number - 1).first()


def make_parent_hash(block):
    parent = find_parent_by_block_number(block)
    return parent and parent.hash or block.default_parent_hash


class BaseWalletFactory(factory.django.DjangoModelFactory):
    address = factory.Faker("ethereum_address")

    class Meta:
        model = BaseEthereumAccount
        django_get_or_create = ("address",)


class ChainFactory(factory.django.DjangoModelFactory):
    id = TEST_CHAIN_ID
    highest_block = 0
    provider = factory.RelatedFactory(
        "hub20.apps.blockchain.factories.base.Web3ProviderFactory", factory_related_name="chain"
    )
    native_token = factory.RelatedFactory(
        "hub20.apps.blockchain.factories.base.NativeTokenFactory", factory_related_name="chain"
    )

    @factory.post_generation
    def blocks(obj, create, extracted, **kw):
        if not create:
            return

        if not extracted:
            BlockFactory(number=obj.highest_block, chain=obj)
        else:
            for block in extracted:
                obj.blocks.add(block)

    class Meta:
        model = Chain
        django_get_or_create = ("id",)


class SyncedChainFactory(ChainFactory):
    highest_block = START_BLOCK_NUMBER
    provider = factory.RelatedFactory(
        "hub20.apps.blockchain.factories.base.SyncedWeb3ProviderFactory",
        factory_related_name="chain",
    )


class BlockFactory(factory.django.DjangoModelFactory):
    chain = factory.SubFactory(SyncedChainFactory)
    hash = factory.Faker("hex64")
    number = factory.Sequence(lambda n: n + START_BLOCK_NUMBER)
    timestamp = factory.LazyAttribute(lambda obj: timezone.now())
    parent_hash = factory.LazyAttribute(lambda obj: make_parent_hash(obj))
    uncle_hashes: List = []

    class Params:
        default_parent_hash = factory.Faker("hex64")

    class Meta:
        model = Block


class TransactionDataFactory(factory.django.DjangoModelFactory):
    hash = factory.Faker("hex64")
    chain = factory.SubFactory(SyncedChainFactory)
    from_address = factory.Faker("ethereum_address")
    to_address = factory.Faker("ethereum_address")

    data = factory.LazyAttribute(
        lambda obj: {
            "from": obj.from_address,
            "to": obj.to_address,
            "chainId": hex(obj.chain.id),
            "gasUsed": obj.gas_used,
            "gasPrice": obj.gas_price,
            "blockNumber": obj.block_number,
            "blockHash": obj.block_hash,
            "value": obj.value,
            "nonce": obj.nonce,
        }
    )

    class Meta:
        model = TransactionDataRecord
        django_get_or_create = ("chain",)

    class Params:
        value = 0
        nonce = factory.Sequence(lambda n: n)
        status = 1
        gas_used = 21000
        gas_price = factory.fuzzy.FuzzyInteger(1e9, 100e9)
        block_number = None
        block_hash = None


class TransactionFactory(factory.django.DjangoModelFactory):
    block = factory.SubFactory(BlockFactory)
    hash = factory.Faker("hex64")
    from_address = factory.Faker("ethereum_address")
    to_address = factory.Faker("ethereum_address")

    receipt = factory.LazyAttribute(
        lambda obj: {
            "from": obj.from_address,
            "to": obj.to_address,
            "status": obj.status,
            "blockNumber": obj.block.number,
            "blockHash": obj.block.hash,
            "gasUsed": obj.gas_used,
            "effectiveGasPrice": obj.gas_price,
        }
    )

    class Meta:
        model = Transaction

    class Params:
        status = 1
        gas_used = factory.fuzzy.FuzzyInteger(21000, 200000)
        gas_price = factory.fuzzy.FuzzyInteger(1e9, 100e9)


class NativeTokenFactory(factory.django.DjangoModelFactory):
    chain = factory.SubFactory(ChainFactory)
    name = factory.Sequence(lambda n: f"Native Token #{n:02n}")
    symbol = factory.Sequence(lambda n: f"ETH{n:02n}")

    class Meta:
        model = NativeToken
        django_get_or_create = ("chain",)


class Web3ProviderFactory(factory.django.DjangoModelFactory):
    chain = factory.SubFactory(ChainFactory)
    url = factory.Sequence(lambda n: f"https://web3-{n:02}.example.com")
    connected = True
    is_active = True
    synced = False

    class Meta:

        model = Web3Provider
        django_get_or_create = ("chain",)


class SyncedWeb3ProviderFactory(Web3ProviderFactory):
    chain = factory.SubFactory(SyncedChainFactory)
    synced = True



import logging
from statistics import StatisticsError, mean

from django.conf import settings
from django.core.cache import cache
from qr3.qr import CappedCollection
from redis.connection import parse_url

logger = logging.getLogger(__name__)

BLOCK_HISTORY_COLLECTION_SIZE = 200

CACHE_CONNECTION_PARAMS = parse_url(settings.CACHE_LOCATION)


class MaxPriorityFeeTracker:
    CACHE_KEY_TEMPLATE = "hub20:blockchain:analytics:MAX_PRIORITY_FEE_{chain_id}"

    def _get_cache_key(self, chain_id):
        return self.CACHE_KEY_TEMPLATE.format(chain_id=chain_id)

    def get(self, chain_id):
        return cache.get(self._get_cache_key(chain_id), None)

    def set(self, chain_id, value):
        return cache.set(self._get_cache_key(chain_id), value)


def recommended_eip1559_gas_price(block_data, max_priority_fee=None):
    # Formula follows the simple heuristic described on
    # https://www.blocknative.com/blog/eip-1559-fees. We might want to improve on this later

    try:
        return (2 * block_data.baseFeePerGas) + (max_priority_fee or 0)
    except AttributeError:
        raise ValueError("block does not contain EIP-1559 'baseFeePerGas' field")


def mean_gas_price(block_data):
    try:
        return mean(
            [
                p
                for p in [getattr(t, "gasPrice", None) for t in block_data.transactions]
                if p is not None
            ]
        )
    except StatisticsError:
        return None


def get_historical_block_data(chain_id):
    return CappedCollection(
        f"Blocks:{chain_id}", BLOCK_HISTORY_COLLECTION_SIZE, **CACHE_CONNECTION_PARAMS
    )


def estimate_gas_price(chain_id):
    historical_data = get_historical_block_data(chain_id)
    blocks = historical_data.elements()

    if not blocks:
        return None

    most_recent_block = blocks[0]

    if getattr(most_recent_block, "baseFeePerGas", None):
        # Supports EIP-1559.
        return recommended_eip1559_gas_price(
            most_recent_block, max_priority_fee=MAX_PRIORITY_FEE_TRACKER.get(chain_id)
        )

    # No support for EIP-1559, so we will look at the gas price of all
    # transactions in the N blocks and take a weighted average by the blocks age

    price_history = enumerate([mean_gas_price(block) for block in blocks[::-1]])

    weighted_prices = [age * price for (age, price) in price_history if price is not None]
    denominator = (len(blocks) + 1) * (len(blocks) / 2)

    return int(sum(weighted_prices) / denominator)


MAX_PRIORITY_FEE_TRACKER = MaxPriorityFeeTracker()



import pytest

from hub20.apps.blockchain.choices import ETHEREUM_CHAINS
from hub20.apps.raiden.client.blockchain import get_service_token_address


@pytest.mark.parametrize(
    "chain_id,address",
    [
        (ETHEREUM_CHAINS.mainnet, "0x255Aa6DF07540Cb5d3d297f0D0D4D84cb52bc8e6"),
        (ETHEREUM_CHAINS.goerli, "0x5Fc523e13fBAc2140F056AD7A96De2cC0C4Cc63A"),
    ],
)
def test_can_get_rdn_address_from_contract(chain_id, address):
    assert get_service_token_address(chain_id) == address


__all__ = [
    "test_can_get_rdn_address_from_contract",
]



import logging
from typing import Optional

from django.contrib.sessions.models import Session
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone

from hub20.apps.blockchain.models import (
    BaseEthereumAccount,
    Block,
    Chain,
    Transaction,
    TransactionDataRecord,
)
from hub20.apps.blockchain.signals import block_sealed
from hub20.apps.core import tasks
from hub20.apps.core.choices import PAYMENT_NETWORKS
from hub20.apps.core.consumers import Events
from hub20.apps.core.models import (
    BlockchainPayment,
    BlockchainPaymentRoute,
    Checkout,
    Deposit,
    InternalPayment,
    Payment,
    PaymentConfirmation,
    PaymentOrder,
    RaidenPayment,
    RaidenPaymentRoute,
)
from hub20.apps.core.settings import app_settings
from hub20.apps.core.signals import payment_received
from hub20.apps.ethereum_money.signals import incoming_transfer_broadcast, incoming_transfer_mined
from hub20.apps.raiden.models import Payment as RaidenNodePayment, Raiden
from hub20.apps.raiden.signals import raiden_payment_received
from hub20.apps.wallet import get_wallet_model

logger = logging.getLogger(__name__)
Wallet = get_wallet_model()


def _get_user_id(session: Session) -> Optional[int]:
    try:
        return int(session.get_decoded()["_auth_user_id"])
    except (KeyError, ValueError, TypeError):
        return None


def _get_user_session_keys(user_id):
    now = timezone.now()
    sessions = Session.objects.filter(expire_date__gt=now)
    return [s.session_key for s in sessions if _get_user_id(s) == user_id]


def _check_for_blockchain_payment_confirmations(block_number):
    confirmed_block = block_number - app_settings.Payment.minimum_confirmations

    unconfirmed_payments = BlockchainPayment.objects.filter(
        confirmation__isnull=True, transaction__block__number__lte=confirmed_block
    )

    for payment in unconfirmed_payments:
        PaymentConfirmation.objects.create(payment=payment)


def _publish_block_created_event(chain_id, block_number):
    for checkout in Checkout.objects.with_blockchain_route(block_number):
        logger.debug(
            f"Scheduling publish event for checkout {checkout.id}: block #{block_number} created"
        )
        tasks.publish_checkout_event.delay(
            checkout.id,
            event=Events.BLOCKCHAIN_BLOCK_CREATED.value,
            block=block_number,
            chain_id=chain_id,
        )


@receiver(post_save, sender=Chain)
def on_chain_updated_check_payment_confirmations(sender, **kw):
    chain = kw["instance"]
    _check_for_blockchain_payment_confirmations(chain.highest_block)


@receiver(incoming_transfer_mined, sender=Transaction)
def on_incoming_transfer_mined_check_blockchain_payments(sender, **kw):
    account = kw["account"]
    amount = kw["amount"]
    transaction = kw["transaction"]

    if BlockchainPayment.objects.filter(transaction=transaction).exists():
        logger.info(f"Transaction {transaction} is already recorded for payment")
        return

    route = BlockchainPaymentRoute.objects.filter(
        deposit__currency=amount.currency,
        account=account,
        payment_window__contains=transaction.block.number,
    ).first()

    if not route:
        return

    payment = BlockchainPayment.objects.create(
        route=route,
        amount=amount.amount,
        currency=amount.currency,
        transaction=transaction,
    )
    payment_received.send(sender=BlockchainPayment, payment=payment)


@receiver(incoming_transfer_broadcast, sender=TransactionDataRecord)
def on_incoming_transfer_broadcast_send_notification_to_active_sessions(sender, **kw):
    recipient = kw["account"]
    payment_amount = kw["amount"]
    tx_data = kw["transaction_data"]

    route = BlockchainPaymentRoute.objects.open().filter(account=recipient).first()

    if not route:
        return

    deposit = Deposit.objects.with_blockchain_route().filter(routes=route).first()

    if deposit and deposit.session_key:
        tasks.send_session_event.delay(
            deposit.session_key,
            event=Events.BLOCKCHAIN_DEPOSIT_BROADCAST.value,
            deposit_id=str(deposit.id),
            amount=str(payment_amount.amount),
            token=payment_amount.currency.address,
            transaction=tx_data.hash.hex(),
        )


@receiver(incoming_transfer_broadcast, sender=TransactionDataRecord)
def on_incoming_transfer_broadcast_send_notification_to_open_checkouts(sender, **kw):
    recipient = kw["account"]
    payment_amount = kw["amount"]
    tx_data = kw["transaction_data"]

    route = BlockchainPaymentRoute.objects.open().filter(account=recipient).first()

    if not route:
        return

    checkout = Checkout.objects.with_blockchain_route().filter(routes=route).first()

    if checkout:
        tasks.publish_checkout_event.delay(
            checkout.id,
            event=Events.BLOCKCHAIN_DEPOSIT_BROADCAST.value,
            amount=str(payment_amount.amount),
            token=payment_amount.currency.address,
            transaction=tx_data.hash,
        )


@receiver(raiden_payment_received, sender=RaidenNodePayment)
def on_raiden_payment_received_check_raiden_payments(sender, **kw):
    raiden_payment = kw["payment"]

    if RaidenPayment.objects.filter(payment=raiden_payment).exists():
        logger.info(f"Payment {raiden_payment} is already recorded")
        return

    payment_route = RaidenPaymentRoute.objects.filter(
        identifier=raiden_payment.identifier,
        raiden=raiden_payment.channel.raiden,
    ).first()

    if payment_route is not None:
        amount = raiden_payment.as_token_amount
        RaidenPayment.objects.create(
            route=payment_route,
            amount=amount.amount,
            currency=raiden_payment.token,
            payment=raiden_payment,
        )


@receiver(post_save, sender=Deposit)
@receiver(post_save, sender=PaymentOrder)
@receiver(post_save, sender=Checkout)
def on_order_created_set_blockchain_route(sender, **kw):

    if not kw["created"]:
        return

    deposit = kw["instance"]
    chain = deposit.currency.chain
    chain.refresh_from_db()
    if chain.synced:
        payment_window = BlockchainPaymentRoute.calculate_payment_window(chain)

        busy_routes = BlockchainPaymentRoute.objects.open().filter(
            deposit__currency=deposit.currency
        )
        available_accounts = BaseEthereumAccount.objects.exclude(blockchain_routes__in=busy_routes)

        account = available_accounts.order_by("?").first() or Wallet.generate()

        BlockchainPaymentRoute.objects.create(
            account=account, deposit=deposit, chain=chain, payment_window=payment_window
        )
    else:
        logger.warning("Failed to create blockchain route. Chain data not synced")


@receiver(post_save, sender=Deposit)
@receiver(post_save, sender=PaymentOrder)
@receiver(post_save, sender=Checkout)
def on_order_created_set_raiden_route(sender, **kw):

    if not kw["created"]:
        return

    deposit = kw["instance"]
    raiden = Raiden.objects.first()

    if raiden and raiden.open_channels.filter(token_network__token=deposit.currency).exists():
        raiden.payment_routes.create(deposit=deposit)


@receiver(block_sealed, sender=Block)
def on_block_sealed_publish_block_created_event(sender, **kw):
    block_data = kw["block_data"]
    chain_id = kw["chain_id"]
    _publish_block_created_event(chain_id=chain_id, block_number=block_data.get("number"))


@receiver(block_sealed, sender=Block)
def on_block_sealed_check_confirmed_payments(sender, **kw):
    block_data = kw["block_data"]
    _check_for_blockchain_payment_confirmations(block_data.get("number"))


@receiver(post_save, sender=Block)
def on_block_created_check_confirmed_payments(sender, **kw):
    if kw["created"]:
        block = kw["instance"]
        _check_for_blockchain_payment_confirmations(block.number)


@receiver(post_save, sender=Block)
def on_block_added_publish_expired_blockchain_routes(sender, **kw):
    block = kw["instance"]

    expiring_routes = BlockchainPaymentRoute.objects.filter(
        payment_window__endswith=block.number - 1
    )

    for route in expiring_routes:
        tasks.publish_checkout_event.delay(
            route.deposit_id,
            event=Events.BLOCKCHAIN_ROUTE_EXPIRED.value,
            route=route.account.address,
        )


@receiver(payment_received, sender=BlockchainPayment)
def on_blockchain_payment_received_send_notification(sender, **kw):
    payment = kw["payment"]

    deposit = Deposit.objects.filter(routes__payment=payment).first()

    checkout = Checkout.objects.filter(routes__payment=payment).first()

    payment_data = dict(
        amount=str(payment.amount),
        token=payment.currency.address,
        transaction=payment.transaction.hash,
        block_number=payment.transaction.block.number,
    )

    if deposit and deposit.session_key:
        tasks.send_session_event.delay(
            session_key=deposit.session_key,
            event=Events.BLOCKCHAIN_DEPOSIT_RECEIVED.value,
            deposit_id=str(payment.route.deposit.id),
            **payment_data,
        )

    if checkout:
        tasks.publish_checkout_event.delay(
            checkout.id, event=Events.BLOCKCHAIN_DEPOSIT_RECEIVED.value, **payment_data
        )


@receiver(post_save, sender=RaidenPayment)
@receiver(post_save, sender=InternalPayment)
def on_off_chain_payment_create_confirmation(sender, **kw):
    if kw["created"]:
        PaymentConfirmation.objects.create(payment=kw["instance"])


@receiver(payment_received, sender=BlockchainPayment)
def on_blockchain_payment_received_call_checkout_webhooks(sender, **kw):
    pay = kw["payment"]

    for checkout_id in Checkout.objects.filter(routes__payment=pay).values_list("id", flat=True):
        tasks.call_checkout_webhook.delay(checkout_id)


@receiver(post_save, sender=PaymentConfirmation)
def on_payment_confirmed_call_checkout_webhooks(sender, **kw):
    confirmation = kw["instance"]

    checkouts = Checkout.objects.filter(routes__payment__confirmation=confirmation)
    for checkout_id in checkouts.values_list("id", flat=True):
        tasks.call_checkout_webhook.delay(checkout_id)


@receiver(post_save, sender=PaymentConfirmation)
def on_payment_confirmed_publish_checkout(sender, **kw):
    if not kw["created"]:
        return

    confirmation = kw["instance"]
    payment = Payment.objects.filter(id=confirmation.payment_id).select_subclasses().first()

    if not payment:
        return

    checkouts = Checkout.objects.filter(routes__payment=payment)
    checkout_id = checkouts.values_list("id", flat=True).first()

    if checkout_id is None:
        return

    payment_method = {
        InternalPayment: PAYMENT_NETWORKS.internal,
        BlockchainPayment: PAYMENT_NETWORKS.blockchain,
        RaidenPayment: PAYMENT_NETWORKS.raiden,
    }.get(type(payment))

    event = {
        InternalPayment: Events.INTERNAL_DEPOSIT_CONFIRMED,
        BlockchainPayment: Events.BLOCKCHAIN_DEPOSIT_CONFIRMED,
        RaidenPayment: Events.RAIDEN_DEPOSIT_CONFIRMED,
    }.get(type(payment))

    tasks.publish_checkout_event.delay(
        checkout_id,
        amount=str(payment.amount),
        token=payment.currency.address,
        event=event and event.value,
        payment_method=payment_method,
    )


__all__ = [
    "on_chain_updated_check_payment_confirmations",
    "on_incoming_transfer_mined_check_blockchain_payments",
    "on_incoming_transfer_broadcast_send_notification_to_active_sessions",
    "on_incoming_transfer_broadcast_send_notification_to_open_checkouts",
    "on_raiden_payment_received_check_raiden_payments",
    "on_order_created_set_blockchain_route",
    "on_order_created_set_raiden_route",
    "on_block_added_publish_expired_blockchain_routes",
    "on_block_created_check_confirmed_payments",
    "on_block_sealed_check_confirmed_payments",
    "on_blockchain_payment_received_send_notification",
    "on_blockchain_payment_received_call_checkout_webhooks",
    "on_off_chain_payment_create_confirmation",
    "on_payment_confirmed_call_checkout_webhooks",
    "on_payment_confirmed_publish_checkout",
]



import getpass
import logging

import ethereum
from django.core.management.base import BaseCommand
from eth_utils import to_checksum_address

from hub20.apps.blockchain.client import make_web3
from hub20.apps.blockchain.models import BaseEthereumAccount, Web3Provider
from hub20.apps.ethereum_money.models import EthereumToken, EthereumTokenAmount, KeystoreAccount
from hub20.apps.raiden.client.blockchain import mint_tokens

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Mints tests tokens"

    def add_arguments(self, parser):
        parser.add_argument("-a", "--account", required=True, type=str)
        parser.add_argument("-t", "--token", required=True, type=str)
        parser.add_argument("--amount", default=1000, type=int)
        parser.add_argument("--chain-id", "-c", dest="chain_id", required=True, type=int)

    def handle(self, *args, **options):
        address = to_checksum_address(options["account"])
        accounts = [
            acc
            for acc in BaseEthereumAccount.objects.filter(address=address).select_subclasses()
            if acc.private_key is not None
        ]

        account = accounts and accounts.pop()

        if not account:
            private_key = getpass.getpass(f"Private Key for {address} required: ")
            generated_address = to_checksum_address(ethereum.utils.privtoaddr(private_key.strip()))
            assert generated_address == address, "Private Key does not match"
            account = KeystoreAccount(address=address, private_key=private_key)

        provider = Web3Provider.available.get(chain_id=options["chain_id"])
        w3 = make_web3(provider=provider)

        is_mainnet = provider.chain_id == 1
        if is_mainnet:
            logger.error("Can only mint tokens on testnets")
            return

        token = EthereumToken.make(options["token"], provider.chain)

        amount = EthereumTokenAmount(amount=options["amount"], currency=token)
        mint_tokens(w3=w3, account=account, amount=amount)



# Generated by Django 4.0 on 2022-01-29 00:31

import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import hub20.apps.blockchain.fields
import hub20.apps.ethereum_money.fields
import hub20.apps.raiden.models
import model_utils.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('ethereum_money', '0001_initial'),
        ('blockchain', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Raiden',
            fields=[
                ('baseethereumaccount_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='blockchain.baseethereumaccount')),
                ('url', hub20.apps.raiden.models.RaidenURLField(unique=True)),
                ('web3_provider', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='blockchain.web3provider')),
            ],
            bases=('blockchain.baseethereumaccount',),
        ),
        migrations.CreateModel(
            name='RaidenManagementOrder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('transaction_hash', hub20.apps.blockchain.fields.HexField(max_length=64, null=True, unique=True)),
                ('raiden', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='raiden.raiden')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='auth.user')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TokenNetwork',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('address', hub20.apps.blockchain.fields.EthereumAddressField()),
                ('token', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='ethereum_money.ethereumtoken')),
            ],
        ),
        migrations.CreateModel(
            name='TokenNetworkChannel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', hub20.apps.blockchain.fields.Uint256Field()),
                ('participant_addresses', django.contrib.postgres.fields.ArrayField(base_field=hub20.apps.blockchain.fields.EthereumAddressField(), size=2)),
                ('token_network', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='channels', to='raiden.tokennetwork')),
            ],
        ),
        migrations.CreateModel(
            name='TokenNetworkChannelStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', model_utils.fields.StatusField(choices=[('opened', 'opened'), ('settling', 'settling'), ('settled', 'settled'), ('unusable', 'unusable'), ('closed', 'closed'), ('closing', 'closing')], default='opened', max_length=100, no_check_for_status=True, verbose_name='status')),
                ('status_changed', model_utils.fields.MonitorField(default=django.utils.timezone.now, monitor='status', verbose_name='status changed')),
                ('channel', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='status', to='raiden.tokennetworkchannel')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RaidenManagementOrderResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('order', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='result', to='raiden.raidenmanagementorder')),
                ('transaction', models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, to='blockchain.transaction')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RaidenManagementOrderError',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('message', models.TextField(blank=True, null=True)),
                ('order', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='error', to='raiden.raidenmanagementorder')),
                ('transaction', models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, to='blockchain.transaction')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Channel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', model_utils.fields.StatusField(choices=[('opened', 'opened'), ('settling', 'settling'), ('settled', 'settled'), ('unusable', 'unusable'), ('closed', 'closed'), ('closing', 'closing')], default='opened', max_length=100, no_check_for_status=True, verbose_name='status')),
                ('status_changed', model_utils.fields.MonitorField(default=django.utils.timezone.now, monitor='status', verbose_name='status changed')),
                ('identifier', models.PositiveIntegerField()),
                ('partner_address', hub20.apps.blockchain.fields.EthereumAddressField(db_index=True)),
                ('balance', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('total_deposit', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('total_withdraw', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('raiden', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='channels', to='raiden.raiden')),
                ('token_network', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='raiden.tokennetwork')),
            ],
            options={
                'unique_together': {('raiden', 'token_network', 'identifier'), ('raiden', 'token_network', 'partner_address')},
            },
        ),
        migrations.CreateModel(
            name='UserDepositContractOrder',
            fields=[
                ('raidenmanagementorder_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='raiden.raidenmanagementorder')),
                ('amount', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('chain', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_deposit_orders', to='blockchain.chain')),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.ethereumtoken')),
            ],
            options={
                'abstract': False,
            },
            bases=('raiden.raidenmanagementorder', models.Model),
        ),
        migrations.CreateModel(
            name='TokenNetworkChannelEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=32)),
                ('channel', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='raiden.tokennetworkchannel')),
                ('transaction', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='blockchain.transaction')),
            ],
            options={
                'unique_together': {('channel', 'transaction')},
            },
        ),
        migrations.CreateModel(
            name='Payment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('timestamp', models.DateTimeField()),
                ('identifier', hub20.apps.blockchain.fields.Uint256Field()),
                ('sender_address', hub20.apps.blockchain.fields.EthereumAddressField()),
                ('receiver_address', hub20.apps.blockchain.fields.EthereumAddressField()),
                ('channel', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='payments', to='raiden.channel')),
            ],
            options={
                'unique_together': {('channel', 'identifier', 'sender_address', 'receiver_address')},
            },
        ),
        migrations.CreateModel(
            name='LeaveTokenNetworkOrder',
            fields=[
                ('raidenmanagementorder_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='raiden.raidenmanagementorder')),
                ('token_network', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='raiden.tokennetwork')),
            ],
            options={
                'abstract': False,
            },
            bases=('raiden.raidenmanagementorder',),
        ),
        migrations.CreateModel(
            name='JoinTokenNetworkOrder',
            fields=[
                ('raidenmanagementorder_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='raiden.raidenmanagementorder')),
                ('amount', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('token_network', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='raiden.tokennetwork')),
            ],
            options={
                'abstract': False,
            },
            bases=('raiden.raidenmanagementorder',),
        ),
        migrations.CreateModel(
            name='ChannelWithdrawOrder',
            fields=[
                ('raidenmanagementorder_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='raiden.raidenmanagementorder')),
                ('amount', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('channel', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='raiden.channel')),
            ],
            options={
                'abstract': False,
            },
            bases=('raiden.raidenmanagementorder',),
        ),
        migrations.CreateModel(
            name='ChannelDepositOrder',
            fields=[
                ('raidenmanagementorder_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='raiden.raidenmanagementorder')),
                ('amount', hub20.apps.ethereum_money.fields.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('channel', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='raiden.channel')),
            ],
            options={
                'abstract': False,
            },
            bases=('raiden.raidenmanagementorder',),
        ),
    ]



from raiden_contracts.contract_manager import gas_measurements
from rest_framework import serializers
from rest_framework_nested.relations import NestedHyperlinkedIdentityField
from rest_framework_nested.serializers import NestedHyperlinkedModelSerializer

from hub20.apps.blockchain.client import make_web3
from hub20.apps.blockchain.models import Web3Provider
from hub20.apps.blockchain.serializers import HexadecimalField
from hub20.apps.ethereum_money.models import EthereumTokenAmount
from hub20.apps.ethereum_money.serializers import (
    EthereumTokenSerializer,
    HyperlinkedEthereumTokenSerializer,
    TokenValueField,
)
from hub20.apps.ethereum_money.typing import TokenAmount

from . import models
from .client.blockchain import get_service_token, get_service_token_contract
from .client.node import RaidenClient


class ChainField(serializers.PrimaryKeyRelatedField):
    queryset = models.Chain.objects.filter(tokens__tokennetwork__isnull=False).distinct()


class TokenNetworkField(serializers.RelatedField):
    queryset = models.TokenNetwork.objects.all()
    lookup_field = "address"


class TokenNetworkSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="token-network-detail", lookup_field="address"
    )
    token = HyperlinkedEthereumTokenSerializer()

    class Meta:
        model = models.TokenNetwork
        fields = ("url", "address", "token")
        read_only_fields = ("url", "address", "token")


class ServiceDepositSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="service-deposit-detail")
    raiden = serializers.HyperlinkedRelatedField(
        view_name="raiden-detail", queryset=models.Raiden.objects.all()
    )
    transaction = HexadecimalField(read_only=True, source="result.transaction.hash")
    chain = ChainField()
    token = EthereumTokenSerializer(source="currency", read_only=True)
    amount = TokenValueField()
    error = serializers.CharField(source="error.message", read_only=True)

    def validate_amount(self, value):
        if value <= 0:
            raise serializers.ValidationError("Need to deposit a value larger than 0")

        return value

    def validate(self, data):
        chain = data.pop("chain")
        w3 = make_web3(provider=chain.provider)
        token = get_service_token(w3=w3)
        raiden = data["raiden"]

        if chain != raiden.chain:
            raise serializers.ValidationError(
                f"{raiden} does not seem to be connected to chain {chain.id}"
            )

        contract = get_service_token_contract(w3=w3)
        deposit_amount = EthereumTokenAmount(currency=token, amount=data["amount"])
        balance = token.from_wei(contract.functions.balanceOf(raiden.address).call())

        if balance < deposit_amount:
            raise serializers.ValidationError(f"Insufficient balance: {balance}")

        return data

    def create(self, validated_data):
        request = self.context.get("request")

        chain = validated_data.pop("chain")
        w3 = make_web3(provider=chain.provider)
        token = get_service_token(w3=w3)

        return self.Meta.model.objects.create(user=request.user, currency=token, **validated_data)

    class Meta:
        model = models.UserDepositContractOrder
        fields = ("url", "created", "raiden", "amount", "token", "chain", "transaction", "error")
        read_only_fields = ("url", "created", "raiden", "token", "transaction", "error")


class ChannelSerializer(NestedHyperlinkedModelSerializer):
    url = NestedHyperlinkedIdentityField(
        view_name="raiden-channels-detail",
        parent_lookup_kwargs={
            "raiden_pk": "raiden_id",
        },
    )
    token = HyperlinkedEthereumTokenSerializer(source="token_network.token", read_only=True)

    class Meta:
        model = models.Channel
        fields = ("url", "token", "identifier", "partner_address", "status", "balance")
        read_only_fields = ("url", "token", "identifier", "partner_address", "status", "balance")


class ChannelManagementSerializer(serializers.ModelSerializer):
    channel = serializers.HyperlinkedRelatedField(view_name="channel-detail", read_only=True)
    amount = TokenValueField()

    def create(self, validated_data):
        channel = self.get_channel()
        request = self.context["request"]

        return self.Meta.model.objects.create(
            raiden=channel.raiden, channel=channel, user=request.user, **validated_data
        )

    def get_channel(self):
        view = self.context.get("view")
        return view and view.get_object()


class ChannelDepositSerializer(ChannelManagementSerializer):
    class Meta:
        model = models.ChannelDepositOrder
        fields = ("id", "created", "channel", "amount")
        read_only_fields = ("id", "created", "channel")


class ChannelWithdrawSerializer(ChannelManagementSerializer):
    class Meta:
        model = models.ChannelWithdrawOrder
        fields = ("created", "channel", "amount")
        read_only_fields = ("created", "channel")

    def validate_amount(self, data):
        channel = self.get_channel()
        if channel is None:
            raise serializers.ValidationError("Can not get channel information")

        token = channel.token
        amount = TokenAmount(data).normalize()
        withdraw_amount = EthereumTokenAmount(amount=amount, currency=token)
        channel_balance = EthereumTokenAmount(amount=channel.balance, currency=token)

        if withdraw_amount > channel_balance:
            raise serializers.ValidationError(f"Insufficient balance: {channel_balance.formatted}")

        return data


class RaidenSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="raiden-detail")
    chain = serializers.HyperlinkedRelatedField(
        view_name="blockchain:chain-detail", read_only=True
    )
    channels = ChannelSerializer(many=True)
    status = serializers.HyperlinkedIdentityField(view_name="raiden-status")

    class Meta:
        model = models.Raiden
        fields = read_only_fields = ("url", "status", "chain", "address", "channels")


class RaidenStatusSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="raiden-status")
    raiden = serializers.HyperlinkedIdentityField(view_name="raiden-detail")
    online = serializers.SerializerMethodField()
    cost_estimates = serializers.SerializerMethodField()

    def get_online(self, obj):
        client = RaidenClient(raiden_account=obj)
        return "ready" == client.get_status()

    def get_cost_estimates(self, obj):
        try:
            provider = Web3Provider.active.get(chain_id=obj.chain.id)
            w3 = make_web3(provider=provider)
            gas_price = w3.eth.generate_gas_price()
            gas_costs = gas_measurements()
            actions = {
                "udc-deposit": gas_costs["UserDeposit.deposit"],
                "udc-topup": gas_costs["UserDeposit.deposit (increase balance)"],
                "udc-withdraw": (
                    gas_costs["UserDeposit.planWithdraw"] + gas_costs["UserDeposit.withdraw"]
                ),
                "channel-setup": gas_costs["TokenNetwork.openChannelWithDeposit"],
                "channel-open": gas_costs["TokenNetwork.openChannel"],
                "channel-deposit": gas_costs["TokenNetwork.setTotalDeposit"],
                "channel-withdraw": gas_costs["TokenNetwork.setTotalWithdraw"],
                "channel-close": gas_costs["TokenNetwork.closeChannel"],
            }
            return {k: v * gas_price for k, v in actions.items()}
        except Web3Provider.DoesNotExist:
            return None

    class Meta:
        model = models.Raiden
        fields = read_only_fields = ("url", "raiden", "online", "cost_estimates")


class JoinTokenNetworkOrderSerializer(serializers.ModelSerializer):
    token_network = serializers.HyperlinkedRelatedField(
        view_name="token-network-detail", lookup_field="address", read_only=True
    )

    def get_token_network(self):
        view = self.context.get("view")
        return view and view.get_object()

    def create(self, validated_data):
        request = self.context["request"]
        token_network = self.get_token_network()

        return self.Meta.model.objects.create(
            raiden=self.raiden, token_network=token_network, user=request.user, **validated_data
        )

    def validate(self, attrs):
        attrs = super().validate(attrs)
        token_network = self.get_token_network()

        assert isinstance(self.raiden, models.Raiden)
        if token_network in self.raiden.token_networks:
            raise serializers.ValidationError(
                f"Already joined token network {token_network.address}"
            )

        return attrs

    class Meta:
        model = models.JoinTokenNetworkOrder
        fields = ("id", "created", "token_network", "amount")
        read_only_fields = ("id", "created", "token_network")



import pytest
from django.test import TestCase
from eth_utils import is_checksum_address

from ..factories import EthereumAccountFactory


@pytest.mark.django_db(transaction=True)
class BaseTestCase(TestCase):
    pass


class EthereumAccountTestCase(BaseTestCase):
    def setUp(self):
        self.account = EthereumAccountFactory()

    def test_account_address_is_checksummed(self):
        self.assertTrue(is_checksum_address(self.account.address))


__all__ = ["EthereumAccountTestCase"]



from typing import Any

from django.db import models

from .validators import tokenlist_uri_validator


class TokenlistStandardURLField(models.URLField):
    default_validators = [tokenlist_uri_validator]


class EthereumTokenAmountField(models.DecimalField):
    def __init__(self, *args: Any, **kw: Any) -> None:
        kw.setdefault("decimal_places", 18)
        kw.setdefault("max_digits", 32)

        super().__init__(*args, **kw)



import os
from typing import Optional

import ethereum
from django.db import models
from django.db.models import Max
from hdwallet import HDWallet
from hdwallet.symbols import ETH

from hub20.apps.blockchain.fields import HexField
from hub20.apps.blockchain.models import BaseEthereumAccount

from .app_settings import HD_WALLET_MNEMONIC, HD_WALLET_ROOT_KEY


class ColdWallet(BaseEthereumAccount):
    @classmethod
    def generate(cls):
        raise TypeError("Cold wallets do not store private keys and can not be generated")


class KeystoreAccount(BaseEthereumAccount):
    private_key = HexField(max_length=64, unique=True)

    @classmethod
    def generate(cls):
        private_key = os.urandom(32)
        address = ethereum.utils.privtoaddr(private_key)
        checksum_address = ethereum.utils.checksum_encode(address.hex())
        return cls.objects.create(address=checksum_address, private_key=private_key.hex())


class HierarchicalDeterministicWallet(BaseEthereumAccount):
    BASE_PATH_FORMAT = "m/44'/60'/0'/0/{index}"

    index = models.PositiveIntegerField(unique=True)

    @property
    def private_key(self):
        wallet = self.__class__.get_wallet(index=self.index)
        return wallet.private_key()

    @property
    def private_key_bytes(self) -> bytes:
        return bytearray.fromhex(self.private_key)

    @classmethod
    def get_wallet(cls, index: int) -> HDWallet:
        wallet = HDWallet(symbol=ETH)

        if HD_WALLET_MNEMONIC:
            wallet.from_mnemonic(mnemonic=HD_WALLET_MNEMONIC)
        elif HD_WALLET_ROOT_KEY:
            wallet.from_xprivate_key(xprivate_key=HD_WALLET_ROOT_KEY)
        else:
            raise ValueError("Can not generate new addresses for HD Wallets. No seed available")

        wallet.from_path(cls.BASE_PATH_FORMAT.format(index=index))
        return wallet

    @classmethod
    def generate(cls):
        latest_generation = cls.get_latest_generation()
        index = 0 if latest_generation is None else latest_generation + 1
        wallet = HierarchicalDeterministicWallet.get_wallet(index)
        return cls.objects.create(index=index, address=wallet.p2pkh_address())

    @classmethod
    def get_latest_generation(cls) -> Optional[int]:
        return cls.objects.aggregate(generation=Max("index")).get("generation")


__all__ = [
    "ColdWallet",
    "KeystoreAccount",
    "HierarchicalDeterministicWallet",
]



from rest_framework import permissions


class IsStoreOwnerOrAnonymousReadOnly(permissions.IsAuthenticatedOrReadOnly):
    def has_object_permission(self, request, view, obj):
        is_safe = request.method in permissions.SAFE_METHODS

        return obj.owner == request.user if request.user.is_authenticated else is_safe


class IsStoreOwner(permissions.IsAuthenticated):
    def has_object_permission(self, request, view, obj):
        return request.user.is_authenticated and obj.owner == request.user



from django.contrib import admin

from . import forms, models


@admin.register(models.EthereumToken)
class TokenAdmin(admin.ModelAdmin):
    form = forms.TokenForm

    search_fields = ["symbol", "name", "address", "chain__name"]
    list_display = ["symbol", "name", "address", "chain", "is_listed"]
    list_filter = ["is_listed", "chain"]
    readonly_fields = ["chain", "symbol", "name", "address", "decimals"]


# @admin.register(models.TokenList)
# class TokenListAdmin(admin.ModelAdmin):
#     form = forms.TokenListForm
#     list_display = ["name", "description"]


@admin.register(models.WrappedToken)
class WrappedTokenAdmin(admin.ModelAdmin):
    form = forms.WrappedTokenForm

    list_display = ("wrapped", "wrapper")


@admin.register(models.StableTokenPair)
class StableTokenPairAdmin(admin.ModelAdmin):
    form = forms.StableTokenPairForm

    list_display = ("token", "currency")
    list_filter = ("currency",)



from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import path

urlpatterns = static(settings.STATIC_URL, document_root=settings.STATIC_ROOT) + [
    path("", admin.site.urls)
]



from django.urls import path
from rest_framework.routers import SimpleRouter
from rest_framework_nested.routers import NestedSimpleRouter

from . import views

router = SimpleRouter(trailing_slash=False)
router.register("nodes", views.RaidenViewSet, basename="raiden")

nodes_router = NestedSimpleRouter(router, "nodes", lookup="raiden")
nodes_router.register("channels", views.ChannelViewSet, basename="raiden-channels")
nodes_router.register("connections", views.TokenNetworkViewSet, basename="token-network")

urlpatterns = [
    path(
        "services/deposits",
        views.ServiceDepositListView.as_view(),
        name="service-deposit-list",
    ),
    path(
        "services/deposits/<int:pk>",
        views.ServiceDepositDetailView.as_view(),
        name="service-deposit-detail",
    ),
]

urlpatterns.extend(router.urls)
urlpatterns.extend(nodes_router.urls)



# Generated by Django 4.0 on 2022-01-28 23:27

from django.db import migrations, models
import django.db.models.deletion
import hub20.apps.blockchain.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('blockchain', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ColdWallet',
            fields=[
                ('baseethereumaccount_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='blockchain.baseethereumaccount')),
            ],
            bases=('blockchain.baseethereumaccount',),
        ),
        migrations.CreateModel(
            name='HierarchicalDeterministicWallet',
            fields=[
                ('baseethereumaccount_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='blockchain.baseethereumaccount')),
                ('index', models.PositiveIntegerField(unique=True)),
            ],
            bases=('blockchain.baseethereumaccount',),
        ),
        migrations.CreateModel(
            name='KeystoreAccount',
            fields=[
                ('baseethereumaccount_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='blockchain.baseethereumaccount')),
                ('private_key', hub20.apps.blockchain.fields.HexField(max_length=64, unique=True)),
            ],
            bases=('blockchain.baseethereumaccount',),
        ),
    ]



from decimal import Decimal
from typing import TypeVar, Union

Wei = int
TokenAmount = Decimal

EthereumClient_T = TypeVar("EthereumClient_T")
TokenAmount_T = Union[int, float, Decimal, TokenAmount, Wei]



import logging

from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.dispatch import receiver

logger = logging.getLogger(__name__)
User = get_user_model()

from hub20.apps.core.models import UserPreferences


@receiver(post_save, sender=User)
def on_user_created_create_user_preferences(sender, **kw):
    if kw["created"]:
        UserPreferences.objects.get_or_create(user=kw["instance"])


__all__ = ["on_user_created_create_user_preferences"]



import logging
import sys

from django.core.management.base import BaseCommand
from eth_utils import to_checksum_address

from hub20.apps.blockchain.client import make_web3
from hub20.apps.blockchain.models import Chain, Web3Provider
from hub20.apps.ethereum_money.client import get_token_information
from hub20.apps.ethereum_money.models import EthereumToken

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Gets information about a token from the blockchain and lists it"

    def add_arguments(self, parser):
        parser.add_argument("--chain-id", "-c", dest="chain_id", required=True, type=int)
        parser.add_argument("--address", "-a", dest="address", required=True, type=str)
        parser.add_argument("--listed", action="store_true")

    def handle(self, *args, **options):

        try:
            chain_id = options["chain_id"]
            chain = Chain.objects.get(id=chain_id)
            provider = Web3Provider.objects.get(chain=chain)
        except Chain.DoesNotExist:
            logger.error(f"Chain {chain_id} not found.")
            sys.exit(-1)
        except Web3Provider.DoesNotExist:
            logger.error("No web3 provider for {chain.name}")
            sys.exit(-1)

        w3 = make_web3(provider=provider)

        token_address = to_checksum_address(options["address"])

        logger.info(f"Checking token {token_address}...")
        try:
            token_data = get_token_information(w3=w3, address=token_address)
            EthereumToken.make(
                address=token_address,
                chain=provider.chain,
                is_listed=options["listed"],
                **token_data,
            )
        except OverflowError:
            logger.error(f"{token_address} is not a valid address or not ERC20-compliant")
        except Exception as exc:
            logger.exception(f"Failed to load token data for {token_address}", exc_info=exc)



import logging
from decimal import Decimal

from django.conf import settings
from django.test.signals import setting_changed

logger = logging.getLogger(__name__)


class AppSettings:
    class Transfer:
        minimum_confirmations = 10

    class Payment:
        minimum_confirmations = 5
        blockchain_route_lifetime = 100  # In blocks
        raiden_route_lifetime = 15 * 60  # In seconds

    class Raiden:
        minimum_ether_required = Decimal("0.025")
        minimum_service_token_required = Decimal("5.5")

    def __init__(self):
        self.load()

    def load(self):
        ATTRS = {
            "TRANSFER_MININUM_CONFIRMATIONS": (self.Transfer, "minimum_confirmations"),
            "PAYMENT_MININUM_CONFIRMATIONS": (self.Payment, "minimum_confirmations"),
            "PAYMENT_BLOCKCHAIN_ROUTE_LIFETIME": (self.Payment, "blockchain_route_lifetime"),
            "RAIDEN_MINIMUM_ETHER_REQUIRED": (self.Raiden, "minimum_ether_required"),
            "RAIDEN_MINIMUM_RDN_REQUIRED": (self.Raiden, "minimum_service_token_required"),
        }
        user_settings = getattr(settings, "HUB20", {})

        for setting, value in user_settings.items():
            if setting not in ATTRS:
                logger.warning(f"Ignoring {setting} as it is not a setting for HUB20")
                continue

            setting_class, attr = ATTRS[setting]
            setattr(setting_class, attr, value)


app_settings = AppSettings()


def reload_settings(*args, **kw):
    global app_settings
    setting = kw["setting"]
    if setting == "HUB20":
        app_settings.load()


setting_changed.connect(reload_settings)



import logging

from web3 import Web3

from hub20.apps.blockchain.models import Chain, Web3Provider
from hub20.apps.core.settings import app_settings
from hub20.apps.ethereum_money.client import get_account_balance
from hub20.apps.ethereum_money.models import EthereumToken, EthereumTokenAmount
from hub20.apps.raiden.client.blockchain import get_service_deposit_balance, get_service_token
from hub20.apps.raiden.exceptions import RaidenMissingPrecondition
from hub20.apps.raiden.models import Raiden

logger = logging.getLogger(__name__)


def check_is_ethereum_node_synced(w3: Web3):
    synced = bool(not w3.eth.syncing)
    if not synced:
        raise RaidenMissingPrecondition(f"Not connected to chain #{w3.eth.chain_id}")


def check_required_ether_balance(raiden: Raiden, w3: Web3, chain: Chain):
    native_token = EthereumToken.make_native(chain)
    on_chain_balance = get_account_balance(w3=w3, token=native_token, address=raiden.address)
    required_amount = EthereumTokenAmount(
        amount=app_settings.Raiden.minimum_ether_required, currency=native_token
    )

    if on_chain_balance < required_amount:
        raise RaidenMissingPrecondition(
            f"Minimum balance of {required_amount.formatted} must be available"
        )
    else:
        logger.info(f"{on_chain_balance.formatted} available for transactions")


def check_required_service_token_deposit(raiden: Raiden, w3: Web3):
    service_token = get_service_token(w3=w3)
    required_balance = EthereumTokenAmount(
        amount=app_settings.Raiden.minimum_service_token_required, currency=service_token
    )
    on_chain_balance = get_account_balance(w3=w3, token=service_token, address=raiden.address)
    deposit_balance = get_service_deposit_balance(w3=w3, raiden=raiden)

    logger.info(
        f"Service token: {on_chain_balance.formatted} on chain, {deposit_balance.formatted} locked"
    )
    if deposit_balance < required_balance:

        raise RaidenMissingPrecondition(
            f"Minimum of {required_balance.formatted} must be deposited for Raiden Services"
        )
    else:
        logger.info(f"{deposit_balance.formatted} deposited for payment of Raiden Services")


def ensure_preconditions(raiden: Raiden, w3: Web3):
    chain_id = w3.eth.chain_id
    w3_uri = w3.provider.endpoint_uri
    try:
        provider = Web3Provider.available.get(chain_id=chain_id, url=w3_uri)
    except Web3Provider.DoesNotExist:
        raise RaidenMissingPrecondition(f"Not connected to network {chain_id}")

    check_is_ethereum_node_synced(w3=w3)
    check_required_ether_balance(raiden=raiden, w3=w3, chain=provider.chain)
    check_required_service_token_deposit(raiden=raiden, w3=w3)



from unittest.mock import patch

import pytest
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.test import TestCase

from hub20.apps.blockchain.models import Block
from hub20.apps.blockchain.signals import block_sealed
from hub20.apps.blockchain.tests.mocks import BlockMock
from hub20.apps.core.choices import TRANSFER_STATUS
from hub20.apps.core.factories import (
    CheckoutFactory,
    Erc20TokenPaymentConfirmationFactory,
    Erc20TokenPaymentOrderFactory,
    ExternalTransferFactory,
    InternalTransferFactory,
    StoreFactory,
    UserAccountFactory,
)
from hub20.apps.core.models.accounting import ExternalAddressAccount
from hub20.apps.core.models.payments import (
    BlockchainPayment,
    BlockchainPaymentRoute,
    RaidenPaymentRoute,
)
from hub20.apps.core.models.transfers import TransferCancellation
from hub20.apps.core.settings import app_settings
from hub20.apps.core.tests.unit.mocks import (
    MockBlockchainTransferExecutor,
    MockRaidenTransferExecutor,
    mock_fee_estimation,
)
from hub20.apps.ethereum_money.tests.base import add_eth_to_account, add_token_to_account
from hub20.apps.raiden.factories import (
    ChannelFactory,
    PaymentEventFactory,
    RaidenFactory,
    TokenNetworkFactory,
)
from hub20.apps.wallet.factories import EthereumAccountFactory


@pytest.mark.django_db(transaction=True)
class BaseTestCase(TestCase):
    pass


class BlockchainPaymentTestCase(BaseTestCase):
    def setUp(self):
        self.order = Erc20TokenPaymentOrderFactory()
        self.blockchain_route = BlockchainPaymentRoute.objects.filter(deposit=self.order).first()
        self.chain = self.blockchain_route.chain

    def test_transaction_sets_payment_as_received(self):
        add_token_to_account(self.blockchain_route.account, self.order.as_token_amount)
        self.assertTrue(self.order.is_paid)
        self.assertFalse(self.order.is_confirmed)

    def test_transaction_creates_blockchain_payment(self):
        add_token_to_account(self.blockchain_route.account, self.order.as_token_amount)
        self.assertEqual(self.order.payments.count(), 1)

    def test_can_not_add_same_transaction_twice(self):
        add_token_to_account(self.blockchain_route.account, self.order.as_token_amount)
        self.assertEqual(self.order.payments.count(), 1)
        payment = self.order.payments.select_subclasses().first()
        with self.assertRaises(IntegrityError):
            BlockchainPayment.objects.create(
                transaction=payment.transaction,
                route=payment.route,
                amount=payment.amount,
                currency=payment.currency,
            )

    def test_user_balance_is_updated_on_completed_payment(self):
        tx = add_token_to_account(self.blockchain_route.account, self.order.as_token_amount)

        block_number = tx.block.number + app_settings.Payment.minimum_confirmations
        block_data = BlockMock(number=block_number)
        block_sealed.send(sender=Block, chain_id=tx.block.chain_id, block_data=block_data)

        balance_amount = self.order.user.account.get_balance_token_amount(self.order.currency)
        self.assertEqual(balance_amount, self.order.as_token_amount)


class CheckoutTestCase(BaseTestCase):
    def setUp(self):
        self.checkout = CheckoutFactory()
        self.checkout.store.accepted_token_list.tokens.add(self.checkout.currency)

    def test_checkout_user_and_store_owner_are_the_same(self):
        self.assertEqual(self.checkout.store.owner, self.checkout.user)

    def test_checkout_currency_must_be_accepted_by_store(self):
        self.checkout.clean()

        self.checkout.store.accepted_token_list.tokens.clear()
        with self.assertRaises(ValidationError):
            self.checkout.clean()


class RaidenPaymentTestCase(BaseTestCase):
    def setUp(self):
        token_network = TokenNetworkFactory()

        self.channel = ChannelFactory(token_network=token_network)
        self.order = Erc20TokenPaymentOrderFactory(currency=token_network.token)
        self.raiden_route = RaidenPaymentRoute.objects.filter(deposit=self.order).first()

    def test_order_has_raiden_route(self):
        self.assertIsNotNone(self.raiden_route)

    def test_payment_via_raiden_sets_order_as_paid(self):
        PaymentEventFactory(
            channel=self.channel,
            amount=self.order.amount,
            identifier=self.raiden_route.identifier,
            receiver_address=self.channel.raiden.address,
        )
        self.assertTrue(self.order.is_paid)


class StoreTestCase(BaseTestCase):
    def setUp(self):
        self.store = StoreFactory()

    def test_store_rsa_keys_are_valid_pem(self):
        self.assertIsNotNone(self.store.rsa.pk)
        self.assertTrue(type(self.store.rsa.public_key_pem) is str)
        self.assertTrue(type(self.store.rsa.private_key_pem) is str)

        self.assertTrue(self.store.rsa.public_key_pem.startswith("-----BEGIN PUBLIC KEY-----"))
        self.assertTrue(
            self.store.rsa.private_key_pem.startswith("-----BEGIN RSA PRIVATE KEY-----")
        )


class TransferTestCase(BaseTestCase):
    def setUp(self):
        self.sender_account = UserAccountFactory()
        self.receiver_account = UserAccountFactory()
        self.sender = self.sender_account.user
        self.receiver = self.receiver_account.user

        self.deposit = Erc20TokenPaymentConfirmationFactory(
            payment__route__deposit__user=self.sender,
        )

        self.credit = self.deposit.payment.as_token_amount
        self.wallet = EthereumAccountFactory()
        self.fee_amount = mock_fee_estimation()
        self.chain = self.fee_amount.currency.chain

        self.raiden = RaidenFactory()


class InternalTransferTestCase(TransferTestCase):
    def test_transfers_are_finalized_as_executed(self):
        transfer = InternalTransferFactory(
            sender=self.sender,
            receiver=self.receiver,
            currency=self.credit.currency,
            amount=self.credit.amount,
        )

        transfer.execute()
        self.assertTrue(transfer.is_finalized)
        self.assertEqual(transfer.status, TRANSFER_STATUS.executed)
        self.assertTrue(transfer.is_executed)

    def test_transfers_change_balance(self):
        transfer = InternalTransferFactory(
            sender=self.sender,
            receiver=self.receiver,
            currency=self.credit.currency,
            amount=self.credit.amount,
        )

        transfer.execute()
        self.assertTrue(transfer.is_finalized)

        sender_balance = self.sender_account.get_balance_token_amount(self.credit.currency)
        receiver_balance = self.receiver_account.get_balance_token_amount(self.credit.currency)

        self.assertEqual(sender_balance.amount, 0)
        self.assertEqual(receiver_balance, self.credit)

    def test_transfers_fail_with_low_sender_balance(self):
        transfer = InternalTransferFactory(
            sender=self.sender,
            receiver=self.receiver,
            currency=self.credit.currency,
            amount=2 * self.credit.amount,
        )

        transfer.execute()
        self.assertTrue(transfer.is_finalized)
        self.assertEqual(transfer.status, TRANSFER_STATUS.failed)


class ExternalTransferTestCase(TransferTestCase):
    def setUp(self):
        super().setUp()
        add_token_to_account(self.wallet, self.credit)
        add_eth_to_account(self.wallet, self.fee_amount)

        self.transfer = ExternalTransferFactory(
            sender=self.sender, currency=self.credit.currency, amount=self.credit.amount
        )
        self.transfer.EXECUTORS = (MockBlockchainTransferExecutor,)

    @patch.object(MockBlockchainTransferExecutor, "select_for_transfer")
    def test_external_transfers_fail_without_funds(self, select_for_transfer):
        select_for_transfer.return_value = None
        self.transfer.execute()
        self.assertTrue(self.transfer.is_failed)
        self.assertEqual(self.transfer.status, TRANSFER_STATUS.failed)

    @patch.object(MockBlockchainTransferExecutor, "select_for_transfer")
    def test_transfers_can_be_executed_with_enough_balance(self, select_for_transfer):
        select_for_transfer.return_value = MockBlockchainTransferExecutor(self.wallet)
        self.transfer.execute()
        self.assertTrue(self.transfer.is_executed)
        self.assertEqual(self.transfer.status, TRANSFER_STATUS.executed)


class TransferAccountingTestcase(TransferTestCase):
    def test_cancelled_transfer_generate_refunds(self):
        transfer = InternalTransferFactory(
            sender=self.sender,
            receiver=self.receiver,
            currency=self.credit.currency,
            amount=self.credit.amount,
        )
        treasury = transfer.currency.chain.treasury
        cancellation = TransferCancellation.objects.create(
            transfer=transfer, canceled_by=self.sender
        )

        sender_balance_amount = self.sender.account.get_balance_token_amount(
            token=self.credit.currency
        )
        self.assertEqual(sender_balance_amount, self.credit)
        last_treasury_debit = treasury.debits.last()

        self.assertEqual(last_treasury_debit.reference, cancellation)

    def test_external_transfers_generate_accounting_entries_for_wallet_and_external_address(self):
        transfer = ExternalTransferFactory(
            sender=self.sender, currency=self.credit.currency, amount=self.credit.amount
        )

        with patch.object(MockBlockchainTransferExecutor, "select_for_transfer") as select:
            select.return_value = MockBlockchainTransferExecutor(self.wallet)
            transfer.EXECUTORS = (MockBlockchainTransferExecutor,)
            transfer.execute()

        transaction = transfer.execution.blockchaintransferexecution.transaction
        transaction_type = ContentType.objects.get_for_model(transaction)

        wallet_account = self.wallet.onchain_account
        external_account = ExternalAddressAccount.objects.filter(address=transfer.address).first()

        self.assertIsNotNone(external_account)

        external_credit = external_account.credits.filter(reference_type=transaction_type).last()
        wallet_debit = wallet_account.debits.filter(reference_type=transaction_type).last()

        self.assertIsNotNone(wallet_debit)
        self.assertIsNotNone(external_credit)

        self.assertEqual(wallet_debit.as_token_amount, external_credit.as_token_amount)

    def test_blockchain_transfers_create_fee_entries(self):
        transfer = ExternalTransferFactory(
            sender=self.sender, currency=self.credit.currency, amount=self.credit.amount
        )

        with patch.object(MockBlockchainTransferExecutor, "select_for_transfer") as select:
            select.return_value = MockBlockchainTransferExecutor(self.wallet)
            transfer.EXECUTORS = (MockBlockchainTransferExecutor,)
            transfer.execute()

        self.assertTrue(hasattr(transfer, "execution"))
        self.assertTrue(hasattr(transfer.execution, "blockchaintransferexecution"))

        transaction = transfer.execution.blockchaintransferexecution.transaction
        transaction_type = ContentType.objects.get_for_model(transaction)
        ETH = transfer.execution.blockchaintransferexecution.fee.currency

        treasury_book = self.chain.treasury.get_book(token=ETH)
        fee_account = ExternalAddressAccount.get_transaction_fee_account()
        wallet_book = self.wallet.onchain_account.get_book(token=ETH)
        sender_book = transfer.sender.account.get_book(token=ETH)

        entry_filters = dict(reference_type=transaction_type, reference_id=transaction.id)

        self.assertIsNotNone(treasury_book.credits.filter(**entry_filters).last())
        self.assertIsNotNone(wallet_book.debits.filter(**entry_filters).last())
        self.assertIsNotNone(fee_account.credits.filter(**entry_filters).last())
        self.assertIsNotNone(sender_book.debits.filter(**entry_filters).last())

    def test_raiden_transfers_create_entries_for_account_and_external_address(self):
        transfer = ExternalTransferFactory(
            sender=self.sender,
            currency=self.credit.currency,
            amount=self.credit.amount,
            identifier="0xdeadbeef",
        )

        with patch.object(MockRaidenTransferExecutor, "select_for_transfer") as select:
            select.return_value = MockRaidenTransferExecutor(account=self.raiden)
            transfer.EXECUTORS = (MockRaidenTransferExecutor,)
            transfer.execute()

        self.assertTrue(hasattr(transfer, "execution"))
        self.assertTrue(hasattr(transfer.execution, "raidentransferexecution"))
        self.assertIsNotNone(transfer.execution.raidentransferexecution.payment)

        payment = transfer.execution.raidentransferexecution.payment
        transfer_type = ContentType.objects.get_for_model(transfer)

        self.assertEqual(payment.receiver_address, transfer.address)

        raiden_account = payment.channel.raiden.raiden_account
        external_address_account, _ = ExternalAddressAccount.objects.get_or_create(
            address=transfer.address
        )

        transfer_filter = dict(reference_type=transfer_type, reference_id=transfer.id)

        self.assertIsNotNone(raiden_account.debits.filter(**transfer_filter).last())
        self.assertIsNotNone(external_address_account.credits.filter(**transfer_filter).last())


__all__ = [
    "BlockchainPaymentTestCase",
    "CheckoutTestCase",
    "RaidenPaymentTestCase",
    "StoreTestCase",
    "TransferTestCase",
    "InternalTransferTestCase",
    "ExternalTransferTestCase",
    "TransferAccountingTestcase",
]



from typing import Dict

from django.contrib.auth import get_user_model
from django.db import transaction
from ipware import get_client_ip
from rest_framework import serializers
from rest_framework.reverse import reverse

from hub20.apps.blockchain.serializers import EthereumAddressField, HexadecimalField
from hub20.apps.ethereum_money.models import EthereumToken, EthereumTokenAmount
from hub20.apps.ethereum_money.serializers import (
    EthereumTokenSerializer,
    HyperlinkedRelatedTokenField,
    HyperlinkedTokenIdentityField,
    HyperlinkedTokenMixin,
    TokenValueField,
)

from . import models

User = get_user_model()


class UserRelatedField(serializers.SlugRelatedField):
    queryset = User.objects.filter(is_active=True)

    def __init__(self, *args, **kw):
        kw.setdefault("slug_field", "username")
        super().__init__(*args, **kw)


class StoreAcceptedTokenList(serializers.HyperlinkedRelatedField):
    view_name = "user-tokenlist-detail"

    def get_queryset(self):
        request = self.context.get("request")

        return request.user.token_lists.all()


class UserTokenSelectorField(HyperlinkedRelatedTokenField, HyperlinkedTokenMixin):
    def get_attribute(self, instance):
        return instance


class HyperlinkedBalanceIdentityField(serializers.HyperlinkedIdentityField):
    def __init__(self, *args, **kw):
        kw.setdefault("view_name", "balance-detail")
        super().__init__(*args, **kw)

    def get_url(self, obj, view_name, request, format):
        url_kwargs = {"chain_id": obj.chain_id, "address": obj.address}
        return reverse(view_name, kwargs=url_kwargs, request=request, format=format)


class UserSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="users-detail", lookup_field="username")

    class Meta:
        model = User
        fields = read_only_fields = ("url", "username", "first_name", "last_name", "email")


class TokenBalanceSerializer(EthereumTokenSerializer):
    token = HyperlinkedTokenIdentityField(view_name="token-detail")
    amount = TokenValueField(read_only=True, source="balance")

    class Meta:
        model = EthereumTokenSerializer.Meta.model
        fields = read_only_fields = (
            "token",
            "amount",
        )


class HyperlinkedTokenBalanceSerializer(HyperlinkedTokenMixin, TokenBalanceSerializer):
    url = HyperlinkedTokenIdentityField(view_name="balance-detail")

    view_name = "balance-detail"

    class Meta:
        model = TokenBalanceSerializer.Meta.model
        fields = ("url",) + TokenBalanceSerializer.Meta.fields
        read_only_fields = ("url",) + TokenBalanceSerializer.Meta.read_only_fields


class TransferSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="transfer-detail")
    address = EthereumAddressField(required=False, allow_null=True)
    recipient = UserRelatedField(source="receiver", required=False, allow_null=True)
    token = HyperlinkedRelatedTokenField(source="currency")
    target = serializers.CharField(read_only=True)
    status = serializers.CharField(read_only=True)

    def validate_recipient(self, value):
        request = self.context["request"]
        if value == request.user:
            raise serializers.ValidationError("You can not make a transfer to yourself")
        return value

    def validate(self, data):
        # Check if we have a valid recipient
        address = data.get("address", None)
        recipient = data.get("receiver", None)

        if not address and not recipient:
            raise serializers.ValidationError(
                "Either one of recipient or address must be provided"
            )
        if address and recipient:
            raise serializers.ValidationError(
                "Choose recipient by address or username, but not both at the same time"
            )

        # We do need to check the balance here though because the amount
        # corresponding to the transfer is deducted from the user's balance
        # upon creation for two reasons: keeping the accounting books balanced
        # and ensuring that users can not overdraw.

        # There is also the cost of transfer fees (especially for on-chain
        # transfers), but given that these can not be predicted and the hub
        # operator might waive the fees from the users, we do not do any
        # charging here and deduct fees only after the transaction is complete.
        request = self.context["request"]

        currency = data["currency"]
        transfer_amount = EthereumTokenAmount(currency=currency, amount=data["amount"])
        user_balance_amount = request.user.account.get_balance_token_amount(currency)

        if not user_balance_amount:
            raise serializers.ValidationError("No balance available", code="invalid")

        if user_balance_amount < transfer_amount:
            raise serializers.ValidationError("Insufficient balance", code="insufficient")

        return data

    def create(self, validated_data):
        request = self.context["request"]

        return self.Meta.model.objects.create(sender=request.user, **validated_data)

    class Meta:
        model = models.Transfer
        fields = (
            "id",
            "url",
            "address",
            "recipient",
            "amount",
            "token",
            "memo",
            "identifier",
            "status",
            "target",
        )
        read_only_fields = ("id", "status", "target")


class TransferExecutionSerializer(serializers.ModelSerializer):
    token = HyperlinkedRelatedTokenField(source="transfer.currency")
    target = serializers.CharField(source="transfer.target", read_only=True)
    amount = TokenValueField(source="transfer.amount")

    class Meta:
        model = models.TransferExecution
        fields = read_only_fields = ("created", "token", "amount", "target")


class PaymentRouteSerializer(serializers.ModelSerializer):
    type = serializers.CharField(source="name", read_only=True)

    @staticmethod
    def get_serializer_class(route):
        return {
            models.InternalPaymentRoute: InternalPaymentRouteSerializer,
            models.BlockchainPaymentRoute: BlockchainPaymentRouteSerializer,
            models.RaidenPaymentRoute: RaidenPaymentRouteSerializer,
        }.get(type(route), PaymentRouteSerializer)


class InternalPaymentRouteSerializer(PaymentRouteSerializer):
    class Meta:
        model = models.InternalPaymentRoute
        fields = read_only_fields = ("recipient", "type")


class BlockchainPaymentRouteSerializer(PaymentRouteSerializer):
    address = EthereumAddressField(source="account.address", read_only=True)
    network_id = serializers.IntegerField(source="order.chain_id", read_only=True)
    start_block = serializers.IntegerField(source="start_block_number", read_only=True)
    expiration_block = serializers.IntegerField(source="expiration_block_number", read_only=True)

    class Meta:
        model = models.BlockchainPaymentRoute
        fields = read_only_fields = (
            "address",
            "network_id",
            "start_block",
            "expiration_block",
            "type",
        )


class RaidenPaymentRouteSerializer(PaymentRouteSerializer):
    address = EthereumAddressField(source="raiden.address", read_only=True)

    class Meta:
        model = models.RaidenPaymentRoute
        fields = read_only_fields = ("address", "expiration_time", "identifier", "type")


class PaymentSerializer(serializers.ModelSerializer):
    id = serializers.UUIDField()
    url = serializers.HyperlinkedIdentityField(view_name="payments-detail")
    currency = EthereumTokenSerializer()
    confirmed = serializers.BooleanField(source="is_confirmed", read_only=True)

    class Meta:
        model = models.Payment
        fields = read_only_fields = (
            "id",
            "url",
            "created",
            "currency",
            "amount",
            "confirmed",
        )


class InternalPaymentSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.InternalPayment
        fields = PaymentSerializer.Meta.fields + ("identifier", "user", "memo")
        read_only_fields = PaymentSerializer.Meta.read_only_fields + (
            "identifier",
            "user",
            "memo",
        )


class BlockchainPaymentSerializer(PaymentSerializer):
    transaction = HexadecimalField(source="transaction.hash", read_only=True)
    block = serializers.IntegerField(source="transaction.block.number", read_only=True)

    class Meta:
        model = models.BlockchainPayment
        fields = PaymentSerializer.Meta.fields + ("identifier", "transaction", "block")
        read_only_fields = PaymentSerializer.Meta.read_only_fields + (
            "identifier",
            "transaction",
            "block",
        )


class RaidenPaymentSerializer(PaymentSerializer):
    raiden = serializers.CharField(source="payment.channel.raiden.address")

    class Meta:
        model = models.RaidenPayment
        fields = PaymentSerializer.Meta.fields + ("identifier", "raiden")
        read_only_fields = PaymentSerializer.Meta.read_only_fields + (
            "identifier",
            "raiden",
        )


class DepositSerializer(serializers.ModelSerializer):

    token = HyperlinkedRelatedTokenField(source="currency")
    routes = serializers.SerializerMethodField()
    payments = serializers.SerializerMethodField()

    def create(self, validated_data: Dict):
        request = self.context["request"]
        with transaction.atomic():
            return self.Meta.model.objects.create(
                user=request.user, session_key=request.session.session_key, **validated_data
            )

    def get_routes(self, obj):
        def get_route_serializer(route):
            serializer_class = PaymentRouteSerializer.get_serializer_class(route)
            return serializer_class(route, context=self.context)

        return [get_route_serializer(route).data for route in obj.routes.select_subclasses()]

    def get_payments(self, obj):
        def get_payment_serializer(payment):
            return {
                models.InternalPayment: InternalPaymentSerializer,
                models.BlockchainPayment: BlockchainPaymentSerializer,
                models.RaidenPayment: RaidenPaymentSerializer,
            }.get(type(payment), PaymentSerializer)(payment, context=self.context)

        return [get_payment_serializer(payment).data for payment in obj.payments]

    class Meta:
        model = models.Deposit
        fields = (
            "id",
            "token",
            "created",
            "routes",
            "payments",
            "status",
        )
        read_only_fields = ("id", "created", "status")


class HttpDepositSerializer(DepositSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="deposit-detail")

    class Meta:
        model = DepositSerializer.Meta.model
        fields = ("url",) + DepositSerializer.Meta.fields
        read_only_fields = DepositSerializer.Meta.read_only_fields


class PaymentOrderSerializer(DepositSerializer):
    amount = TokenValueField()

    class Meta:
        model = models.PaymentOrder
        fields = DepositSerializer.Meta.fields + ("amount",)
        read_only_fields = DepositSerializer.Meta.read_only_fields


class HttpPaymentOrderSerializer(PaymentOrderSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="payment-order-detail")

    class Meta:
        model = PaymentOrderSerializer.Meta.model
        fields = PaymentOrderSerializer.Meta.fields + ("url",)
        read_only_fields = PaymentOrderSerializer.Meta.read_only_fields


class PaymentOrderReadSerializer(PaymentOrderSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="payment-order-detail")

    class Meta:
        model = PaymentOrderSerializer.Meta.model
        fields = PaymentOrderSerializer.Meta.fields + ("url",)
        read_only_fields = PaymentOrderSerializer.Meta.read_only_fields


class PaymentConfirmationSerializer(serializers.ModelSerializer):
    token = HyperlinkedRelatedTokenField(source="payment.currency")
    amount = TokenValueField(source="payment.amount")
    route = serializers.SerializerMethodField()

    def get_route(self, obj):
        route = models.PaymentRoute.objects.get_subclass(id=obj.payment.route_id)
        serializer_class = PaymentRouteSerializer.get_serializer_class(route)
        return serializer_class(route, context=self.context).data

    class Meta:
        model = models.PaymentConfirmation
        fields = read_only_fields = ("created", "token", "amount", "route")


class CheckoutSerializer(PaymentOrderSerializer):
    store = serializers.PrimaryKeyRelatedField(queryset=models.Store.objects.all())

    def validate(self, data):
        currency = data["currency"]
        store = data["store"]
        if currency not in store.accepted_currencies.all():
            raise serializers.ValidationError(f"{currency.symbol} is not accepted at {store.name}")

        return data

    def create(self, validated_data):
        request = self.context.get("request")
        client_ip, _ = get_client_ip(request)
        store = validated_data.pop("store")

        with transaction.atomic():
            return models.Checkout.objects.create(
                store=store,
                user=store.owner,
                session_key=request.session.session_key,
                requester_ip=client_ip,
                **validated_data,
            )

    class Meta:
        model = models.Checkout
        fields = PaymentOrderSerializer.Meta.fields + (
            "store",
            "external_identifier",
            "voucher",
        )
        read_only_fields = PaymentOrderSerializer.Meta.read_only_fields + ("voucher",)


class HttpCheckoutSerializer(CheckoutSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="checkout-detail")

    class Meta:
        model = models.Checkout
        fields = ("url",) + CheckoutSerializer.Meta.fields
        read_only_fields = CheckoutSerializer.Meta.read_only_fields


class StoreSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="store-detail")
    site_url = serializers.URLField(source="url")
    public_key = serializers.CharField(source="rsa.public_key_pem", read_only=True)

    class Meta:
        model = models.Store
        fields = ("id", "url", "name", "site_url", "public_key")
        read_only_fields = ("id", "public_key")


class StoreViewerSerializer(StoreSerializer):
    accepted_currencies = HyperlinkedRelatedTokenField(many=True)

    class Meta:
        model = models.Store
        fields = read_only_fields = (
            "id",
            "url",
            "name",
            "site_url",
            "public_key",
            "accepted_currencies",
        )


class StoreEditorSerializer(StoreSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="user-store-detail")
    accepted_token_list = StoreAcceptedTokenList()
    checkout_webhook_url = serializers.URLField(allow_null=True)

    def create(self, validated_data):
        request = self.context.get("request")
        return models.Store.objects.create(owner=request.user, **validated_data)

    class Meta:
        model = models.Store
        fields = StoreSerializer.Meta.fields + (
            "accepted_token_list",
            "checkout_webhook_url",
        )
        read_only_fields = StoreSerializer.Meta.read_only_fields


class UserTokenSerializer(EthereumTokenSerializer):
    url = HyperlinkedTokenIdentityField(view_name="user-token-detail")
    token = HyperlinkedTokenIdentityField(view_name="token-detail", read_only=True)
    address = EthereumAddressField(read_only=True)
    chain_id = serializers.PrimaryKeyRelatedField(read_only=True)

    class Meta:
        model = EthereumTokenSerializer.Meta.model
        fields = ("url", "token") + EthereumTokenSerializer.Meta.fields
        read_only_fields = ("url", "token") + EthereumTokenSerializer.Meta.fields


class UserTokenCreatorSerializer(serializers.ModelSerializer):
    url = HyperlinkedTokenIdentityField(view_name="user-token-detail")
    token = UserTokenSelectorField(view_name="token-detail")

    def validate(self, data):
        token = data["token"]
        if not EthereumToken.tradeable.filter(id=token.id).exists():
            raise serializers.ValidationError(
                f"Token {token.symbol} on chain #{token.chain_id} is not listed for trade"
            )
        return data

    def create(self, validated_data):
        request = self.context["request"]
        token = validated_data["token"]
        native_token = EthereumToken.make_native(token.chain)
        request.user.preferences.tokens.add(token)
        request.user.preferences.tokens.add(native_token)
        return token

    class Meta:
        model = EthereumTokenSerializer.Meta.model
        fields = ("url", "token") + EthereumTokenSerializer.Meta.fields
        read_only_fields = ("url",) + EthereumTokenSerializer.Meta.fields


class BookEntrySerializer(serializers.ModelSerializer):
    amount = serializers.CharField(source="as_token_amount")
    reference_type = serializers.CharField(source="reference_type.model")
    reference = serializers.SerializerMethodField()
    type = serializers.SerializerMethodField()

    def get_type(self, obj):
        return obj.__class__.__name__.lower()

    def get_summary(self, obj):
        return {
            models.Transfer: "transfer",
            models.TransferExecution: "transfer sent",
            models.PaymentConfirmation: "payment received",
        }.get(type(obj.reference))

    def get_reference(self, obj):
        params = {
            models.TransferExecution: lambda: {
                "viewname": "transfer-detail",
                "kwargs": {"pk": obj.reference.transfer.pk},
            },
            models.PaymentConfirmation: lambda: {
                "viewname": "payments-detail",
                "kwargs": {"pk": obj.reference.payment.pk},
            },
            models.Transfer: lambda: {
                "viewname": "transfer-detail",
                "kwargs": {"pk": obj.reference.pk},
            },
        }.get(type(obj.reference))

        return params and reverse(request=self.context.get("request"), **params())

    class Meta:
        read_only_fields = fields = (
            "id",
            "created",
            "amount",
            "type",
            "reference_type",
            "reference",
        )


class CreditSerializer(BookEntrySerializer):
    class Meta:
        model = models.Credit
        fields = read_only_fields = BookEntrySerializer.Meta.fields


class DebitSerializer(BookEntrySerializer):
    class Meta:
        model = models.Debit
        fields = read_only_fields = BookEntrySerializer.Meta.fields


class UserPreferencesSerializer(serializers.ModelSerializer):
    tokens = HyperlinkedRelatedTokenField(
        view_name="token-detail",
        queryset=EthereumToken.tradeable.all(),
        many=True,
    )

    class Meta:
        model = models.UserPreferences
        fields = ("tokens",)


class AccountingBookSerializer(serializers.Serializer):
    token = HyperlinkedTokenIdentityField(view_name="token-detail", source="*")
    total_credit = TokenValueField(read_only=True)
    total_debit = TokenValueField(read_only=True)
    balance = TokenValueField(read_only=True)


class WalletBalanceSheetSerializer(serializers.ModelSerializer):
    url = serializers.SerializerMethodField()
    address = EthereumAddressField(source="account.address", read_only=True)
    balances = AccountingBookSerializer(many=True, read_only=True)

    def get_url(self, obj):
        return reverse(
            "accounting-wallets-detail",
            kwargs=dict(address=obj.account.address),
            request=self.context["request"],
        )

    class Meta:
        model = models.WalletAccount
        fields = read_only_fields = ("url", "address", "balances")



import factory
from django.contrib.contenttypes.models import ContentType

from hub20.apps.blockchain.factories import EthereumProvider
from hub20.apps.core import models
from hub20.apps.ethereum_money.factories import (
    Erc20TokenValueModelFactory,
    EthereumTokenValueModelFactory,
)

from .base import UserFactory

factory.Faker.add_provider(EthereumProvider)


class EthereumBookEntryFactory(EthereumTokenValueModelFactory):
    class Meta:
        model = models.BookEntry


class Erc20TokenBookEntryFactory(Erc20TokenValueModelFactory):
    class Meta:
        model = models.BookEntry


class UserAccountFactory(factory.django.DjangoModelFactory):
    user = factory.SubFactory(UserFactory)

    class Meta:
        model = models.UserAccount
        django_get_or_create = ("user",)


class UserBookFactory(factory.django.DjangoModelFactory):
    owner_id = factory.SelfAttribute("owner.id")
    owner_type = factory.LazyAttribute(lambda o: ContentType.objects.get_for_model(o.owner))
    owner = factory.SubFactory(UserAccountFactory)

    class Meta:
        model = models.Book


class EthereumCreditFactory(EthereumBookEntryFactory):
    book = factory.SubFactory(UserBookFactory)

    class Meta:
        model = models.Credit


class Erc20TokenCreditFactory(EthereumBookEntryFactory):
    book = factory.SubFactory(UserBookFactory)

    class Meta:
        model = models.Credit


class EthereumDebitFactory(EthereumBookEntryFactory):
    book = factory.SubFactory(UserBookFactory)

    class Meta:
        model = models.Debit


class Erc20TokenDebitFactory(EthereumBookEntryFactory):
    book = factory.SubFactory(UserBookFactory)

    class Meta:
        model = models.Debit


__all__ = [
    "EthereumBookEntryFactory",
    "EthereumCreditFactory",
    "EthereumDebitFactory",
    "Erc20TokenCreditFactory",
    "Erc20TokenDebitFactory",
    "Erc20TokenBookEntryFactory",
    "UserAccountFactory",
]



from django import forms

from hub20.apps.blockchain.models import Web3Provider
from hub20.apps.blockchain.validators import uri_parsable_scheme_validator

from . import models

raiden_url_validator = uri_parsable_scheme_validator(("http", "https"))


class RaidenURLField(forms.URLField):
    default_validators = [raiden_url_validator]


class RaidenForm(forms.ModelForm):
    url = RaidenURLField()
    web3_provider = forms.ModelChoiceField(queryset=Web3Provider.available.all())

    class Meta:
        model = models.Raiden
        fields = ("url", "web3_provider")



from typing import Callable

from dj_rest_auth.views import (
    LoginView,
    LogoutView,
    PasswordChangeView,
    PasswordResetConfirmView,
    PasswordResetView,
    UserDetailsView,
)
from django.conf import settings
from django.conf.urls.static import static
from django.urls import include, path
from django.views.generic import TemplateView
from drf_yasg import openapi
from drf_yasg.views import get_schema_view
from rest_framework.permissions import AllowAny

from hub20.apps.core.api import urlpatterns as core_urlpatterns

from .views import IndexView


def make_auth_view(url_path: str, view_class, view_name: str) -> Callable:
    return path(url_path, view_class.as_view(), name=view_name)


urlpatterns = [
    # URLs that do not require a session or valid token
    make_auth_view("accounts/password/reset", PasswordResetView, "rest_password_reset"),
    make_auth_view(
        "accounts/password/confirm",
        PasswordResetConfirmView,
        "rest_password_reset_confirm",
    ),
    make_auth_view("accounts/password/change", PasswordChangeView, "rest_password_change"),
    make_auth_view("session/login", LoginView, "rest_login"),
    make_auth_view("session/logout", LogoutView, "rest_logout"),
    path("my/profile", UserDetailsView.as_view(), name="rest_user_details"),
    path("register/", include("dj_rest_auth.registration.urls")),
    path("networks/blockchains/", include("hub20.apps.blockchain.api", namespace="blockchain")),
    path("networks/raiden/", include("hub20.apps.raiden.api")),
    path("", IndexView.as_view(), name="index"),
]

urlpatterns.extend(core_urlpatterns)

if settings.SERVE_OPENAPI_URLS:
    schema_view = get_schema_view(
        openapi.Info(
            title="Hub20",
            description="REST API - Description",
            default_version="1.0.0",
        ),
        public=True,
        permission_classes=(AllowAny,),
    )

    urlpatterns.extend(
        [
            path(
                "docs",
                TemplateView.as_view(
                    template_name="swagger-ui.html", extra_context={"schema_url": "openapi-schema"}
                ),
                name="swagger-ui",
            ),
            path(
                "openapi",
                schema_view.without_ui(),
                name="openapi-schema",
            ),
        ]
    )

urlpatterns.extend(static(settings.STATIC_URL, document_root=settings.STATIC_ROOT))



import logging
from typing import Optional

from django.contrib.auth import get_user_model
from eth_utils import to_checksum_address
from ethereum.abi import ContractTranslator
from web3 import Web3
from web3._utils.events import get_event_data
from web3.exceptions import LogTopicError, MismatchedABI

from hub20.apps.blockchain.client import make_web3
from hub20.apps.blockchain.factories.base import FAKER
from hub20.apps.blockchain.models import BaseEthereumAccount, Chain
from hub20.apps.blockchain.typing import Address, EthereumAccount_T
from hub20.apps.ethereum_money.abi import EIP20_ABI
from hub20.apps.ethereum_money.app_settings import TRANSFER_GAS_LIMIT
from hub20.apps.ethereum_money.models import EthereumToken, EthereumTokenAmount
from hub20.apps.ethereum_money.typing import EthereumClient_T

logger = logging.getLogger(__name__)
User = get_user_model()


def get_transaction_events(transaction_receipt, event_abi):
    w3 = Web3()
    events = []
    for log in transaction_receipt.logs:
        try:
            event = get_event_data(w3.codec, event_abi, log)
            events.append(event)
        except (MismatchedABI, LogTopicError) as exc:
            tx_hash = transaction_receipt.transactionHash.hex()
            logger.info(f"{exc} when getting events from tx {tx_hash}")
    return events


def encode_transfer_data(recipient_address, amount: EthereumTokenAmount):
    translator = ContractTranslator(EIP20_ABI)
    encoded_data = translator.encode_function_call("transfer", (recipient_address, amount.as_wei))
    return f"0x{encoded_data.hex()}"


def get_transfer_gas_estimate(w3: Web3, token: EthereumToken):
    if token.is_ERC20:
        contract = w3.eth.contract(abi=EIP20_ABI, address=token.address)
        return contract.functions.transfer(FAKER.ethereum_address(), 0).estimateGas(
            {"from": FAKER.ethereum_address()}
        )
    else:
        return 21000


def get_estimate_fee(w3: Web3, token: EthereumToken) -> EthereumTokenAmount:
    native_token = EthereumToken.make_native(chain=token.chain)

    gas_price = w3.eth.generateGasPrice()
    gas_estimate = get_transfer_gas_estimate(w3=w3, token=token)
    return native_token.from_wei(gas_estimate * gas_price)


def get_max_fee(w3: Web3) -> EthereumTokenAmount:
    chain = Chain.active.get(id=w3.eth.chain_id)
    native_token = EthereumToken.make_native(chain=chain)

    gas_price = chain.gas_price_estimate or w3.eth.generateGasPrice()
    return native_token.from_wei(TRANSFER_GAS_LIMIT * gas_price)


def get_account_balance(w3: Web3, token: EthereumToken, address: Address) -> EthereumTokenAmount:
    if token.is_ERC20:
        contract = w3.eth.contract(abi=EIP20_ABI, address=token.address)
        return token.from_wei(contract.functions.balanceOf(address).call())
    else:
        return token.from_wei(w3.eth.getBalance(address))


def get_token_information(w3: Web3, address):
    contract = w3.eth.contract(abi=EIP20_ABI, address=to_checksum_address(address))
    return {
        "name": contract.functions.name().call(),
        "symbol": contract.functions.symbol().call(),
        "decimals": contract.functions.decimals().call(),
    }


def make_token(w3: Web3, address) -> EthereumToken:
    token_data = get_token_information(w3=w3, address=address)
    chain = Chain.active.get(id=w3.eth.chain_id)
    return EthereumToken.make(chain=chain, address=address, **token_data)


class EthereumClient:
    def __init__(self, account: EthereumAccount_T, w3: Optional[Web3] = None) -> None:
        self.account = account

    def build_transfer_transaction(self, recipient, amount: EthereumTokenAmount):
        token = amount.currency

        w3 = make_web3(provider=token.chain.provider)
        chain_id = w3.eth.chain_id
        message = f"Connected to network {chain_id}, token {token.symbol} is on {token.chain_id}"
        assert token.chain_id == chain_id, message

        transaction_params = {
            "chainId": chain_id,
            "nonce": w3.eth.getTransactionCount(self.account.address),
            "gasPrice": w3.eth.generateGasPrice(),
            "gas": TRANSFER_GAS_LIMIT,
            "from": self.account.address,
        }

        if token.is_ERC20:
            transaction_params.update(
                {"to": token.address, "value": 0, "data": encode_transfer_data(recipient, amount)}
            )
        else:
            transaction_params.update({"to": recipient, "value": amount.as_wei})
        return transaction_params

    def transfer(self, amount: EthereumTokenAmount, address, *args, **kw):
        w3 = make_web3(provider=amount.currency.chain.provider)
        transaction_data = self.build_transfer_transaction(recipient=address, amount=amount)
        signed_tx = self.sign_transaction(transaction_data=transaction_data, w3=w3)
        return w3.eth.sendRawTransaction(signed_tx.rawTransaction)

    def sign_transaction(self, transaction_data, w3: Web3, *args, **kw):
        if not hasattr(self.account, "private_key"):
            raise NotImplementedError("Can not sign transaction without the private key")
        return w3.eth.account.signTransaction(transaction_data, self.account.private_key)

    def get_balance(self, token: EthereumToken):
        w3 = make_web3(provider=token.chain.provider)
        return get_account_balance(w3=w3, token=token, address=self.account.address)

    @classmethod
    def select_for_transfer(
        cls,
        amount: EthereumTokenAmount,
        receiver: Optional[User] = None,
        address: Optional[Address] = None,
    ) -> Optional[EthereumClient_T]:

        chain = amount.currency.chain
        w3 = make_web3(provider=chain.provider)

        transfer_fee: EthereumTokenAmount = cls.estimate_transfer_fees(w3=w3)
        assert transfer_fee.is_native_token

        ETH = transfer_fee.currency

        accounts = BaseEthereumAccount.objects.all().order_by("?")

        if amount.is_native_token:
            amount += transfer_fee

        for account in accounts:
            get_balance = lambda t: get_account_balance(w3=w3, token=t, address=account.address)

            eth_balance = get_balance(ETH)
            token_balance = eth_balance if amount.is_native_token else get_balance(amount.currency)

            if eth_balance >= transfer_fee and token_balance >= amount:
                return cls(account=account, w3=w3)
        return None

    @classmethod
    def estimate_transfer_fees(cls, *args, **kw) -> EthereumTokenAmount:
        w3 = kw["w3"]
        return get_max_fee(w3=w3)


__all__ = [
    "get_transaction_events",
    "encode_transfer_data",
    "get_transfer_gas_estimate",
    "get_estimate_fee",
    "get_max_fee",
    "get_account_balance",
    "get_token_information",
    "make_token",
    "EthereumClient",
]



import datetime
import uuid
from typing import Optional

import jwt
from Crypto.PublicKey import RSA
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models

from hub20.apps.ethereum_money.models import EthereumToken, UserTokenList

from .payments import PaymentOrder


class Store(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    owner = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    name = models.CharField(max_length=300)
    url = models.URLField(help_text="URL for your store public site or information page")
    checkout_webhook_url = models.URLField(null=True, help_text="URL to receive checkout updates")
    accepted_token_list = models.ForeignKey(
        UserTokenList,
        null=True,
        help_text="The list of tokens that will be accepted for payment",
        on_delete=models.SET_NULL,
    )

    @property
    def accepted_currencies(self):
        if self.accepted_token_list:
            return self.accepted_token_list.tokens.all()
        return EthereumToken.tradeable.all()

    def issue_jwt(self, **data):
        data.update(
            {
                "iat": datetime.datetime.utcnow(),
                "iss": str(self.id),
            }
        )

        private_key = self.rsa.private_key_pem.encode()
        return jwt.encode(data, private_key, algorithm="RS256")

    def __str__(self):
        return f"{self.name} ({self.url})"


class StoreRSAKeyPair(models.Model):
    DEFAULT_KEY_SIZE = 2048

    store = models.OneToOneField(Store, on_delete=models.CASCADE, related_name="rsa")
    public_key_pem = models.TextField()
    private_key_pem = models.TextField()

    @classmethod
    def generate(cls, store: Store, keysize: Optional[int] = None):
        bits = keysize or cls.DEFAULT_KEY_SIZE
        key = RSA.generate(bits)
        public_key_pem = key.publickey().export_key().decode()
        private_key_pem = key.export_key().decode()

        pair, _ = cls.objects.update_or_create(
            store=store,
            defaults={"public_key_pem": public_key_pem, "private_key_pem": private_key_pem},
        )
        return pair


class Checkout(PaymentOrder):
    store = models.ForeignKey(Store, on_delete=models.CASCADE)
    external_identifier = models.TextField()
    requester_ip = models.GenericIPAddressField(null=True)

    @property
    def voucher_data(self):
        return {
            "checkout_id": str(self.id),
            "external_identifier": self.external_identifier,
            "token": {"symbol": self.currency.symbol, "address": self.currency.address},
            "payments": [
                {
                    "id": str(p.id),
                    "amount": str(p.amount),
                    "confirmed": p.is_confirmed,
                    "identifier": p.identifier,
                    "route": p.route.get_route_name(),
                }
                for p in self.payments
            ],
            "total_amount": str(self.amount),
            "total_confirmed": str(self.total_confirmed),
            "is_paid": self.is_paid,
            "is_confirmed": self.is_confirmed,
        }

    @property
    def voucher(self):
        return self.store.issue_jwt(**self.voucher_data)

    def clean(self):
        if self.store.owner != self.user:
            raise ValidationError("Creator of payment order must be the same as store owner")

        if self.currency not in self.store.accepted_token_list.tokens.all():
            raise ValidationError(f"{self.store.name} does not accept payment in {self.currency}")


__all__ = ["Store", "StoreRSAKeyPair", "Checkout"]



import asyncio
import logging

import celery_pubsub
from asgiref.sync import sync_to_async
from django.core.cache import cache
from web3 import Web3

from hub20.apps.blockchain.app_settings import BLOCK_SCAN_RANGE
from hub20.apps.blockchain.client import BLOCK_CREATION_INTERVAL, make_web3
from hub20.apps.blockchain.models import EventIndexer, Web3Provider
from hub20.apps.ethereum_money.abi import EIP20_ABI

logger = logging.getLogger(__name__)


async def process_token_transfer_events():
    indexer_name = "ethereum_money:token_transfers"
    while True:
        providers = await sync_to_async(list)(Web3Provider.available.select_related("chain"))
        for provider in providers:
            chain = provider.chain
            event_indexer = await sync_to_async(EventIndexer.make)(chain.id, indexer_name)

            w3: Web3 = make_web3(provider=provider)
            contract = w3.eth.contract(abi=EIP20_ABI)
            current_block = w3.eth.block_number
            last_processed = event_indexer.last_block

            to_block = min(current_block, last_processed + BLOCK_SCAN_RANGE)
            filter_params = dict(fromBlock=last_processed, toBlock=to_block)

            for transfer_event in contract.events.Transfer().getLogs(filter_params):
                await sync_to_async(celery_pubsub.publish)(
                    "blockchain.event.token_transfer.mined",
                    chain_id=w3.eth.chain_id,
                    event_data=transfer_event,
                    provider_url=provider.url,
                )
            event_indexer.last_block = to_block
            await sync_to_async(event_indexer.save)()
        await asyncio.sleep(BLOCK_CREATION_INTERVAL)


async def process_pending_token_transfers():
    CACHE_KEY = "PENDING_TOKEN_TRANSFERS"

    while True:
        providers = await sync_to_async(list)(Web3Provider.available.select_related("chain"))
        for provider in providers:
            w3: Web3 = make_web3(provider=provider)
            contract = w3.eth.contract(abi=EIP20_ABI)

            try:
                for transfer_event in contract.events.Transfer().getLogs({"fromBlock": "pending"}):
                    tx_hash = transfer_event.transactionHash.hex()
                    key = f"{CACHE_KEY}:{tx_hash}"

                    if await sync_to_async(cache.get)(key):
                        logger.debug(f"Event for tx {tx_hash} has already been published")
                        continue

                    await sync_to_async(celery_pubsub.publish)(
                        "blockchain.event.token_transfer.broadcast",
                        chain_id=w3.eth.chain_id,
                        event_data=transfer_event,
                        provider_url=provider.url,
                    )
                    await sync_to_async(cache.set)(key, True, timeout=BLOCK_CREATION_INTERVAL * 2)
            except ValueError:
                logger.warning(f"Can not get transfer logs from {provider.hostname}")
        await asyncio.sleep(1)


__all__ = ["process_token_transfer_events", "process_pending_token_transfers"]



from rest_framework.decorators import action
from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from .models import Chain
from .serializers import ChainSerializer, ChainStatusSerializer


class ChainViewSet(GenericViewSet, ListModelMixin, RetrieveModelMixin):
    serializer_class = ChainSerializer
    queryset = Chain.active.all()

    @action(detail=True, methods=["get"])
    def status(self, request, pk=None):
        chain = self.get_object()
        serializer = ChainStatusSerializer(chain, context={"request": request})
        return Response(serializer.data)
